#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
from logging import INFO
from jinja2.nativetypes import NativeEnvironment

from octobot_commons.logging.logging_util import get_logger, set_global_logger_level
from octobot_tentacles_manager.constants import TENTACLE_TEMPLATE_PATH, \
    TENTACLE_TEMPLATE_DESCRIPTION, TENTACLE_TEMPLATE_EXT, TENTACLE_TEMPLATE_PRE_EXT, \
    TENTACLES_PATH, TENTACLE_CONFIG_TEMPLATE_PRE_EXT, CONFIG_EXT, \
    TENTACLE_CONFIG, TENTACLES_FOLDERS_ARCH

# TODO: remove from to .coveragerc when adapted


class TentacleCreator:
    def __init__(self, config):
        self.config = config
        self.templates = {}
        self.config_templates = {}
        self.logger = get_logger(self.__class__.__name__)
        set_global_logger_level(INFO)

    @staticmethod
    def get_template_path(name):
        return os.path.join(os.path.dirname(os.path.abspath(__file__)), TENTACLE_TEMPLATE_PATH,
                            f"{name}{TENTACLE_TEMPLATE_PRE_EXT}{TENTACLE_TEMPLATE_EXT}")

    @staticmethod
    def get_config_template_path(name):
        return os.path.join(os.path.dirname(os.path.abspath(__file__)), TENTACLE_TEMPLATE_PATH,
                            f"{name}{TENTACLE_CONFIG_TEMPLATE_PRE_EXT}{TENTACLE_TEMPLATE_EXT}")

    def get_templates(self):
        return self.templates

    def get_config_templates(self):
        return self.config_templates

    def load_templates(self):
        self.templates["Description"] = open(self.get_template_path(TENTACLE_TEMPLATE_DESCRIPTION), "r").read()
        # Todo: handle tentacle sub-types
        raise NotImplementedError("Todo: handle tentacle sub-types")
        for tentacle_type in []:
            try:
                self.templates[tentacle_type] = open(self.get_template_path(tentacle_type), "r").read()
            except FileNotFoundError:
                pass

            try:
                self.config_templates[tentacle_type] = open(self.get_config_template_path(tentacle_type), "r").read()
            except FileNotFoundError:
                pass

    def parse_commands(self, commands) -> int:
        command_help = ""
        for tentacle_type in TENTACLES_FOLDERS_ARCH:
            command_help += f"- {tentacle_type}: Create a new {tentacle_type} tentacle\n"

        if commands:
            if commands[0] == "help":
                self.logger.info(f"Welcome in Tentacle Creator, commands are:\n{command_help}")
            else:
                self.load_templates()
                self.logger.warning("TENTACLE CREATOR IS IN DEVELOPMENT")
                for command in commands:
                    self.create_tentacle(command)
            return 0
        else:
            arguments_help = "-c: activates the tentacle creators."
            self.logger.error(f"Invalid arguments, arguments are: {arguments_help}")
            return 1

    def create_tentacle(self, tentacle_type):
        if tentacle_type in TENTACLES_FOLDERS_ARCH:
            try:
                new_tentacle = CreatedTentacle(self.config, tentacle_type, self)
                new_tentacle.ask_description(tentacle_type)
                new_tentacle.create_file()
                new_tentacle.create_config_file()
                self.logger.info(
                    f"{new_tentacle.get_name()} tentacle successfully created in {new_tentacle.get_path()}")
            except Exception as e:
                self.logger.error(f"Tentacle creation failed : {e}")
        else:
            self.logger.warning(f"This tentacle type '{tentacle_type}' does not exist. "
                                f"Tentacle types are: {list(TENTACLES_FOLDERS_ARCH.keys())}")


class CreatedTentacle:
    DEFAULT_TENTACLE_VERSION = "1.0.0"

    def __init__(self, config, tentacle_type, tentacle_creator):
        self.config = config
        self.tentacle_creator = tentacle_creator

        self.header_separator = '"""\n'
        self.t_type = tentacle_type
        self.subtype = ""
        self.name = ""
        self.version = self.DEFAULT_TENTACLE_VERSION
        self.requirements = []
        self.tests = []

        self.config_file = self.get_config_path() if self.subtype in self.tentacle_creator.get_config_templates() else[]
        self.logger = get_logger(self.__class__.__name__)
        set_global_logger_level(INFO)

    def get_path(self):
        return f"{TENTACLES_PATH}/{self.t_type}/{self.subtype}/{self.name}.py"

    def get_config_path(self):
        return f"{TENTACLES_PATH}/{self.t_type}/{self.subtype}/{TENTACLE_CONFIG}/{self.name}{CONFIG_EXT}"

    def get_name(self):
        return self.name

    def ask_description(self, tentacle_type):
        self.name = input(f"Enter your new {self.t_type} tentacle name : ")
        while self.subtype == "":
            sub_types = TENTACLES_FOLDERS_ARCH[tentacle_type]
            if len(sub_types) > 1:
                new_subtype = input(f"Choose your tentacle type in {sub_types} : ")
                if new_subtype in sub_types:
                    self.subtype = new_subtype
                else:
                    self.logger.warning("Invalid tentacle type")
            else:
                self.subtype = sub_types[0]

    def create_file(self):
        try:
            desc_template = NativeEnvironment().from_string(self.tentacle_creator.get_templates()["Description"])
            impl_template = NativeEnvironment().from_string(self.tentacle_creator.get_templates()[self.subtype])
            if not os.path.isfile(self.get_path()):
                with open(self.get_path(), "w") as tentacle_file:
                    tentacle_file.write(self.header_separator)
                    tentacle_file.write(desc_template.render(name=self.name,
                                                             big_name=self.name.title(),
                                                             t_type=self.t_type,
                                                             subtype=self.subtype,
                                                             version=self.version,
                                                             requirements=self.requirements,
                                                             tests=self.tests,
                                                             config=self.config_file))
                    tentacle_file.write("\n"+self.header_separator)
                    tentacle_file.write(impl_template.render(name=self.name,
                                                             big_name=self.name.title(),
                                                             t_type=self.t_type,
                                                             subtype=self.subtype,
                                                             version=self.version,
                                                             requirements=self.requirements,
                                                             tests=self.tests,
                                                             config=self.config_file))

                # TODO add __init__.py management
            else:
                raise Exception("A tentacle with this name already exists")
        except Exception as e:
            raise e

    def create_config_file(self):
        try:
            cfg_template = NativeEnvironment().from_string(self.tentacle_creator.get_config_templates()[self.subtype])
            if not os.path.isfile(self.get_config_path()):
                with open(self.get_config_path(), "w") as config_file:
                    config_file.write(cfg_template.render()[1:])
            else:
                raise Exception("A config with this name already exists")
        except Exception:
            pass



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from os.path import join

from octobot_tentacles_manager.constants import TENTACLES_PATH, TENTACLE_RESOURCES, DOCUMENTATION_EXT
from octobot_tentacles_manager.models.tentacle import Tentacle
from octobot_tentacles_manager.util.tentacle_explorer import load_tentacle_with_metadata

# tentacle_data_by_tentacle_class is used to cache tentacles metadata
_tentacle_by_tentacle_class = None


async def reload_tentacle_by_tentacle_class():
    global _tentacle_by_tentacle_class
    loaded_tentacles = await load_tentacle_with_metadata(TENTACLES_PATH)
    _tentacle_by_tentacle_class = {
        klass: tentacle
        for tentacle in loaded_tentacles
        for klass in tentacle.tentacle_class_names
    }


def get_tentacle_classes() -> dict:
    return _tentacle_by_tentacle_class


def get_resources_path(klass) -> str:
    return join(get_tentacle_module_path(klass), TENTACLE_RESOURCES)


def get_tentacle_module_path(klass) -> str:
    tentacle = get_tentacle(klass)
    return join(tentacle.tentacle_path, tentacle.name)


def get_documentation_file_path(klass) -> str:
    return join(get_resources_path(klass), f"{klass.get_name()}{DOCUMENTATION_EXT}")


def get_tentacle(klass) -> Tentacle:
    try:
        return _tentacle_by_tentacle_class[klass.get_name()]
    except TypeError:
        raise RuntimeError(f"tentacle have not been initialized, call reload_tentacle_data_by_tentacle_class "
                           f"fix this issue")



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from os.path import split

from octobot_tentacles_manager.models.tentacle_factory import TentacleFactory
from octobot_tentacles_manager.models.tentacle_type import TentacleType
from octobot_tentacles_manager.workers.install_worker import InstallWorker


class SingleInstallWorker(InstallWorker):

    def __init__(self, reference_tentacles_dir, tentacle_path,
                 bot_installation_path, use_confirm_prompt, aiohttp_session):
        super().__init__(reference_tentacles_dir, tentacle_path,
                         bot_installation_path, use_confirm_prompt, aiohttp_session)
        self.single_tentacle_path = None
        self.single_tentacle_type = None

    async def process(self, name_filter=None) -> int:
        self.reset_worker()
        if self.single_tentacle_path is None or self.single_tentacle_type is None:
            self.errors.append("Impossible to install tentacle: please provide the tentacle path and type")
        else:
            await self.tentacles_setup_manager.create_missing_tentacles_arch()
            self.progress = 1
            self.total_steps = 1
            split_path = split(self.single_tentacle_path)
            tentacle_name = split_path[1]
            factory = TentacleFactory(split_path[0])
            tentacle = factory.create_tentacle_from_type(tentacle_name, TentacleType(self.single_tentacle_type))
            # remove tentacle type from tentacle origin path since in this context it doesn't exist in filesystem
            tentacle.tentacle_path = split_path[0]
            await tentacle.initialize()
            self.register_to_process_tentacles_modules([tentacle])
            await self._install_tentacle(tentacle)
            await self.tentacles_setup_manager.refresh_user_tentacles_setup_config_file()
            self.tentacles_setup_manager.cleanup_temp_dirs()
            self.logger.info(f"{tentacle_name} installed")
        self.log_summary()
        return len(self.errors)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
from shutil import rmtree
from os import walk, path

from octobot_tentacles_manager.constants import USER_TENTACLE_CONFIG_PATH, \
    TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR, TENTACLES_PATH
from octobot_tentacles_manager.managers.tentacles_setup_manager import TentaclesSetupManager

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

temp_dir = "temp_tests"


async def test_create_missing_tentacles_arch():
    _cleanup()
    tentacles_setup_manager = TentaclesSetupManager(TENTACLES_PATH)
    await tentacles_setup_manager.create_missing_tentacles_arch()
    trading_mode_files_count = sum(1 for _ in walk(TENTACLES_PATH))
    assert trading_mode_files_count == 24
    assert path.exists(USER_TENTACLE_CONFIG_PATH)
    _cleanup()


def _cleanup():
    if path.exists(temp_dir):
        rmtree(temp_dir)
    if path.exists(TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR):
        rmtree(TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR)
    if path.exists(TENTACLES_PATH):
        rmtree(TENTACLES_PATH)
    if path.exists(USER_TENTACLE_CONFIG_PATH):
        rmtree(USER_TENTACLE_CONFIG_PATH)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_tentacles_manager.constants import TENTACLES_PATH
from octobot_tentacles_manager.creators.tentacle_creator import TentacleCreator
from octobot_tentacles_manager.creators.tentacles_package_creator import create_tentacles_package_from_local_tentacles


def start_tentacle_creator(config, commands) -> int:
    tentacle_creator_inst = TentacleCreator(config)
    return tentacle_creator_inst.parse_commands(commands)


async def create_tentacles_package(package_name, tentacles_folder=TENTACLES_PATH,
                                   in_zip=True, with_dev_mode=False) -> int:
    return await create_tentacles_package_from_local_tentacles(package_name, tentacles_folder, in_zip, with_dev_mode)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from os.path import join, isfile

from octobot_tentacles_manager.constants import PYTHON_INIT_FILE
from octobot_tentacles_manager.util.file_util import find_or_create


TENTACLE_IMPORT_HEADER = """from octobot_tentacles_manager.api.inspector import check_tentacle_version
from octobot_commons.logging.logging_util import get_logger
"""

NEW_LINE = "\n"


async def find_or_create_module_init_file(module_root, modules):
    await find_or_create(join(module_root, PYTHON_INIT_FILE), False, get_module_init_file_content(modules))


async def create_tentacle_init_file_if_necessary(tentacle_module_path, tentacle):
    init_file = join(tentacle_module_path, PYTHON_INIT_FILE)
    await find_or_create(init_file, is_directory=False, file_content=_get_default_init_file_content(tentacle))


def update_tentacle_type_init_file(tentacle, target_tentacle_path, remove_import=False):
    # Not async function to avoid conflict between read and write
    init_file = join(target_tentacle_path, PYTHON_INIT_FILE)
    if remove_import:
        # remove import line
        _remove_tentacle_from_tentacle_type_init_file(tentacle, init_file)
    else:
        _add_tentacle_to_tentacle_type_init_file(tentacle, init_file)


def _add_tentacle_to_tentacle_type_init_file(tentacle, init_file):
    init_content = ""
    if isfile(init_file):
        # load import file
        with open(init_file, "r") as init_file_r:
            init_content = init_file_r.read()
    if tentacle.name not in init_content:
        # add import headers if missing
        if TENTACLE_IMPORT_HEADER not in init_content:
            init_content = f"{TENTACLE_IMPORT_HEADER}{init_content}"
        # add import line
        init_content = f"{init_content}{get_tentacle_import_block(tentacle)}"
        with open(init_file, "w+") as init_file_w:
            init_file_w.write(init_content)


def _remove_tentacle_from_tentacle_type_init_file(tentacle, init_file):
    init_content_lines = []
    if isfile(init_file):
        # load import file
        with open(init_file, "r") as init_file_r:
            init_content_lines = init_file_r.readlines()
    if init_content_lines:
        # remove import line
        to_remove_lines_count = len(get_tentacle_import_block(tentacle).split(NEW_LINE)) - 2
        to_remove_start_index = None
        tentacle_name_identifier = f"'{tentacle.name}'"
        for index, line in enumerate(init_content_lines):
            if tentacle_name_identifier in line:
                if index > 1 and init_content_lines[index - 1] == NEW_LINE:
                    # if possible, also remove additional newline
                    to_remove_start_index = index - 1
                    to_remove_lines_count = to_remove_lines_count + 1
                else:
                    to_remove_start_index = index
                break
        if to_remove_start_index is not None:
            del init_content_lines[to_remove_start_index:to_remove_start_index + to_remove_lines_count]
            with open(init_file, "w+") as init_file_w:
                init_file_w.writelines(init_content_lines)


def get_module_init_file_content(modules):
    return NEW_LINE.join(f"from .{module} import *" for module in modules)


def _get_default_init_file_content(tentacle):
    return f"from .{tentacle.name} import {', '.join(tentacle.tentacle_class_names)}"


def _get_single_module_init_line(tentacle):
    return f"from .{tentacle.name} import *"


def get_tentacle_import_block(tentacle):
    return f"""
if check_tentacle_version('{tentacle.version}', '{tentacle.name}', '{tentacle.origin_package}'):
    try:
        {_get_single_module_init_line(tentacle)}
    except Exception as e:
        get_logger('TentacleLoader').exception(e, True, f'Error when loading {tentacle.name}: {{e}}')
"""



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import aiofiles
from os.path import exists

from octobot_commons.config_manager import dump_json


async def read_config(config_file: str) -> dict:
    if exists(config_file):
        async with aiofiles.open(config_file, "r") as config_file_r:
            return json.loads(await config_file_r.read())
    return {}


def sync_read_config(config_file: str) -> dict:
    if exists(config_file):
        with open(config_file, "r") as config_file_r:
            return json.loads(config_file_r.read())
    return {}


async def write_config(config_file: str, content: dict) -> None:
    async with aiofiles.open(config_file, "w+") as config_file_w:
        await config_file_w.write(dump_json(content))


def sync_write_config(config_file: str, content: dict) -> None:
    with open(config_file, "w+") as config_file_w:
        config_file_w.write(dump_json(content))



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from os.path import isfile
from shutil import rmtree

import aiohttp
import pytest
from os import path

from octobot_tentacles_manager.api.installer import install_all_tentacles
from octobot_tentacles_manager.configuration.tentacle_configuration import get_config, update_config, \
    factory_reset_config, get_config_schema_path
from octobot_tentacles_manager.constants import USER_TENTACLE_CONFIG_PATH, TENTACLES_PATH
from octobot_tentacles_manager.loaders.tentacle_loading import reload_tentacle_by_tentacle_class

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_get_config():
    _cleanup()
    async with aiohttp.ClientSession() as session:
        await install_all_tentacles(_tentacles_local_path(), aiohttp_session=session)
    from tentacles.Evaluator.RealTime import InstantFluctuationsEvaluator
    assert get_config(InstantFluctuationsEvaluator) == {
        "price_difference_threshold_percent": 1,
        "volume_difference_threshold_percent": 400
    }
    from tentacles.Services import RedditService
    assert get_config(RedditService) == {}
    _cleanup()


async def test_update_config():
    async with aiohttp.ClientSession() as session:
        await install_all_tentacles(_tentacles_local_path(), aiohttp_session=session)
    from tentacles.Evaluator.RealTime import InstantFluctuationsEvaluator
    config_update = {
        "price_difference_threshold_percent": 2,
        "plop": 42
    }
    update_config(InstantFluctuationsEvaluator, config_update)
    assert get_config(InstantFluctuationsEvaluator) == {
        "price_difference_threshold_percent": 2,
        "volume_difference_threshold_percent": 400,
        "plop": 42
    }
    _cleanup()


async def test_factory_reset_config():
    async with aiohttp.ClientSession() as session:
        await install_all_tentacles(_tentacles_local_path(), aiohttp_session=session)
    from tentacles.Evaluator.RealTime import InstantFluctuationsEvaluator
    config_update = {
        "price_difference_threshold_percent": 2,
        "plop": 42
    }
    update_config(InstantFluctuationsEvaluator, config_update)
    await reload_tentacle_by_tentacle_class()
    factory_reset_config(InstantFluctuationsEvaluator)
    assert get_config(InstantFluctuationsEvaluator) == {
        "price_difference_threshold_percent": 1,
        "volume_difference_threshold_percent": 400
    }
    _cleanup()


async def test_get_config_schema_path():
    async with aiohttp.ClientSession() as session:
        await install_all_tentacles(_tentacles_local_path(), aiohttp_session=session)
    from tentacles.Evaluator.RealTime import InstantFluctuationsEvaluator
    assert isfile(get_config_schema_path(InstantFluctuationsEvaluator))
    _cleanup()


def _tentacles_local_path():
    return path.join("tests", "static", "tentacles.zip")


def _cleanup():
    if path.exists(TENTACLES_PATH):
        rmtree(TENTACLES_PATH)
    if path.exists(USER_TENTACLE_CONFIG_PATH):
        rmtree(USER_TENTACLE_CONFIG_PATH)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from copy import copy
from os import makedirs
from os.path import join, split, exists

from octobot_commons.logging.logging_util import get_logger
from octobot_tentacles_manager.configuration.config_file import read_config, write_config
from octobot_tentacles_manager.constants import USER_TENTACLE_CONFIG_FILE_PATH, DEFAULT_TENTACLE_CONFIG, \
    ACTIVATABLE_TENTACLES, DEFAULT_BOT_PATH
from octobot_tentacles_manager.loaders.tentacle_loading import get_tentacle_classes, reload_tentacle_by_tentacle_class


class TentaclesSetupConfiguration:
    TENTACLE_ACTIVATION_KEY = "tentacle_activation"

    def __init__(self, bot_installation_path=DEFAULT_BOT_PATH, config_path=USER_TENTACLE_CONFIG_FILE_PATH):
        self.logger = get_logger(self.__class__.__name__)
        self.config_path = join(bot_installation_path, config_path)
        self.tentacles_activation = {}

        # TODO: add user tentacle packages (previously stored in config.json as an URL/path list)
        self.user_tentacles = {}

    async def fill_tentacle_config(self, tentacles, default_tentacle_config=DEFAULT_TENTACLE_CONFIG,
                                   remove_missing_tentacles=True):
        activatable_tentacles_in_list = [tentacle_class_name
                                         for tentacle in tentacles
                                         if tentacle.get_simple_tentacle_type() in ACTIVATABLE_TENTACLES
                                         for tentacle_class_name in tentacle.tentacle_class_names]
        await self._update_tentacles_setup_config(activatable_tentacles_in_list,
                                                  default_tentacle_config_file=default_tentacle_config,
                                                  remove_missing_tentacles=remove_missing_tentacles)

    def update_activation_configuration(self, new_config, deactivate_other_evaluators):
        something_changed = False
        for element_name, activated in new_config.items():
            if element_name in self.tentacles_activation:
                current_activation = self.tentacles_activation[element_name]
                if current_activation != activated:
                    self.logger.info(f"Tentacles configuration updated: {element_name} "
                                     f"{'activated' if activated else 'deactivated'}")
                    self.tentacles_activation[element_name] = activated
                    something_changed = True
        if deactivate_other_evaluators:
            something_changed = self._deactivate_other_evaluators(new_config) or something_changed
        return something_changed

    def _deactivate_other_evaluators(self, new_config):
        something_changed = False
        from octobot_commons.tentacles_management.class_inspector import get_class_from_string, \
            evaluator_parent_inspection
        has_evaluators = True
        try:
            from octobot_evaluators.evaluator.TA_evaluator import TAEvaluator
            from octobot_evaluators.evaluator.social_evaluator import SocialEvaluator
            from octobot_evaluators.evaluator.realtime_evaluator import RealTimeEvaluator
            import tentacles.Evaluator.TA as TA
            import tentacles.Evaluator.Social as SE
            import tentacles.Evaluator.RealTime as RE
        except ImportError:
            has_evaluators = False
        for element_name in self.tentacles_activation.keys():
            if element_name not in new_config:
                if self.tentacles_activation[element_name]:
                    is_evaluator = False
                    if has_evaluators:
                        # deactivate only evaluators
                        ta_klass = get_class_from_string(element_name, TAEvaluator,
                                                         TA, evaluator_parent_inspection)
                        se_klass = get_class_from_string(element_name, SocialEvaluator,
                                                         SE, evaluator_parent_inspection)
                        re_klass = get_class_from_string(element_name, RealTimeEvaluator,
                                                         RE, evaluator_parent_inspection)
                        is_evaluator = any(klass is not None
                                           for klass in [ta_klass, se_klass, re_klass])
                    if is_evaluator:
                        self.logger.info(f"Tentacles configuration updated: {element_name} "
                                         f"{'deactivated'}")
                        self.tentacles_activation[element_name] = False
                        something_changed = True
        return something_changed

    def replace_tentacle_activation(self, new_config):
        self.tentacles_activation = copy(new_config)

    async def read_config(self):
        try:
            self._from_dict(await read_config(self.config_path))
        except Exception as e:
            self.logger.error(f"Error when reading tentacles global configuration file ({e}), "
                              "resetting this file with default values. This will not change "
                              "any specific tentacle configuration.")
            if get_tentacle_classes() is None:
                await reload_tentacle_by_tentacle_class()
            activatable_tentacles_in_list = [tentacle_class_name
                                             for tentacle_class_name, tentacle in get_tentacle_classes().items()
                                             if tentacle.get_simple_tentacle_type() in ACTIVATABLE_TENTACLES]
            await self._update_tentacles_setup_config(activatable_tentacles_in_list)
            await self.save_config()

    async def save_config(self):
        parent_dir, _ = split(self.config_path)
        if not exists(parent_dir):
            makedirs(parent_dir)
        await write_config(self.config_path, self._to_dict())

    async def _update_tentacles_setup_config(self,
                                             activatable_tentacles_in_list,
                                             default_tentacle_config_file=DEFAULT_TENTACLE_CONFIG,
                                             remove_missing_tentacles=True):

        default_config = await read_config(default_tentacle_config_file)
        default_activation_config = default_config[self.TENTACLE_ACTIVATION_KEY] \
            if self.TENTACLE_ACTIVATION_KEY in default_config else {}
        for tentacle in activatable_tentacles_in_list:
            self._update_tentacle_activation(tentacle, default_activation_config)
        if remove_missing_tentacles:
            self._filter_tentacle_activation(activatable_tentacles_in_list)

    def _update_tentacle_activation(self, tentacle, default_config):
        if tentacle not in self.tentacles_activation:
            if tentacle in default_config:
                self.tentacles_activation[tentacle] = default_config[tentacle]
            else:
                self.tentacles_activation[tentacle] = False

    def _filter_tentacle_activation(self, activatable_tentacles_in_list):
        for key in list(self.tentacles_activation.keys()):
            if key not in activatable_tentacles_in_list:
                self.tentacles_activation.pop(key)

    def _from_dict(self, input_dict):
        if self.TENTACLE_ACTIVATION_KEY in input_dict:
            self.tentacles_activation = input_dict[self.TENTACLE_ACTIVATION_KEY]

    def _to_dict(self):
        return {
            self.TENTACLE_ACTIVATION_KEY: self.tentacles_activation
        }



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from os.path import join
from shutil import copy

from octobot_tentacles_manager.configuration.config_file import sync_read_config, sync_write_config
from octobot_tentacles_manager.constants import USER_TENTACLE_SPECIFIC_CONFIG_PATH, CONFIG_EXT, CONFIG_SCHEMA_EXT, \
    TENTACLE_CONFIG
from octobot_tentacles_manager.loaders.tentacle_loading import get_tentacle_module_path


def get_config(klass) -> dict:
    return sync_read_config(_get_config_file_path(klass))


def update_config(klass, config_update) -> None:
    config_file = _get_config_file_path(klass)
    current_config = sync_read_config(config_file)
    # only update values in config update not to erase values in root config (might not be editable)
    current_config.update(config_update)
    sync_write_config(config_file, current_config)


def factory_reset_config(klass) -> None:
    reference_config_path = join(_get_reference_config_path(klass), _get_config_file_name(klass))
    copy(reference_config_path, _get_config_file_path(klass))


def get_config_schema_path(klass) -> str:
    return join(_get_reference_config_path(klass), f"{klass.get_name()}{CONFIG_SCHEMA_EXT}")


def _get_config_file_path(klass) -> str:
    return join(USER_TENTACLE_SPECIFIC_CONFIG_PATH, _get_config_file_name(klass))


def _get_reference_config_path(klass) -> str:
    return join(get_tentacle_module_path(klass), TENTACLE_CONFIG)


def _get_config_file_name(klass) -> str:
    return f"{klass.get_name()}{CONFIG_EXT}"



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from asyncio import gather
from os import remove
from os.path import join, exists
from shutil import rmtree

from octobot_commons.logging.logging_util import get_logger
from octobot_tentacles_manager.configuration.tentacles_setup_configuration import TentaclesSetupConfiguration
from octobot_tentacles_manager.constants import USER_TENTACLE_SPECIFIC_CONFIG_PATH, PYTHON_INIT_FILE, \
    TENTACLES_FOLDERS_ARCH, DEFAULT_TENTACLE_CONFIG, TENTACLES_PATH, USER_TENTACLE_CONFIG_PATH, \
    TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR, DEFAULT_BOT_PATH
from octobot_tentacles_manager.managers.tentacles_init_files_manager import find_or_create_module_init_file, \
    get_module_init_file_content
from octobot_tentacles_manager.util.file_util import find_or_create
from octobot_tentacles_manager.util.tentacle_explorer import load_tentacle_with_metadata


class TentaclesSetupManager:

    def __init__(self, tentacle_setup_root_path, bot_installation_path=DEFAULT_BOT_PATH):
        self.tentacle_setup_root_path = tentacle_setup_root_path
        self.default_tentacle_config = DEFAULT_TENTACLE_CONFIG
        self.bot_installation_path = bot_installation_path

    async def refresh_user_tentacles_setup_config_file(self):
        available_tentacle = await load_tentacle_with_metadata(self.tentacle_setup_root_path)
        tentacle_setup_config = TentaclesSetupConfiguration(bot_installation_path=self.bot_installation_path)
        await tentacle_setup_config.read_config()
        await tentacle_setup_config.fill_tentacle_config(available_tentacle, self.default_tentacle_config)
        await tentacle_setup_config.save_config()

    async def create_missing_tentacles_arch(self):
        # tentacle user config folder
        await find_or_create(join(self.bot_installation_path, USER_TENTACLE_SPECIFIC_CONFIG_PATH))
        # tentacles folder
        found_existing_installation = not await find_or_create(self.tentacle_setup_root_path)
        # tentacle main python init file
        await find_or_create(join(self.tentacle_setup_root_path, PYTHON_INIT_FILE), False,
                             get_module_init_file_content(TENTACLES_FOLDERS_ARCH.keys()))
        # tentacle inner architecture
        await TentaclesSetupManager._tentacle_arch_operation(self.tentacle_setup_root_path,
                                                             TENTACLES_FOLDERS_ARCH,
                                                             self._create_missing_files_and_folders,
                                                             find_or_create)
        await self._create_missing_files_and_folders(self.tentacle_setup_root_path, TENTACLES_FOLDERS_ARCH)
        return found_existing_installation

    async def remove_tentacle_arch_init_files(self):
        await TentaclesSetupManager._remove_tentacles_arch_init_file(self.tentacle_setup_root_path, None)
        await TentaclesSetupManager._tentacle_arch_operation(self.tentacle_setup_root_path,
                                                             TENTACLES_FOLDERS_ARCH,
                                                             self._remove_tentacles_arch_init_file,
                                                             self._remove_tentacles_arch_init_file)

    @staticmethod
    def cleanup_temp_dirs():
        if exists(TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR):
            rmtree(TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR)

    @staticmethod
    def is_tentacles_arch_valid(verbose=True, raises=False) -> bool:
        try:
            if not exists(TENTACLES_PATH):
                return False
            import tentacles
            return True
        except (ImportError, SyntaxError) as e:
            if verbose:
                get_logger(TentaclesSetupManager.__name__).exception(e, True, f"Error when importing tentacles: {e}")
            if raises:
                raise e
            return False

    @staticmethod
    def delete_tentacles_arch(force=False, raises=False,
                              with_user_config=False, bot_installation_path=DEFAULT_BOT_PATH):
        if TentaclesSetupManager.is_tentacles_arch_valid(verbose=False, raises=raises) \
          or (force and exists(join(bot_installation_path, TENTACLES_PATH))):
            rmtree(join(bot_installation_path, TENTACLES_PATH))
        if with_user_config and exists(join(bot_installation_path, USER_TENTACLE_CONFIG_PATH)):
            rmtree(join(bot_installation_path, USER_TENTACLE_CONFIG_PATH))

    @staticmethod
    def get_available_tentacles_repos():
        # TODO: add advanced tentacles repos
        return [TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR]

    @staticmethod
    async def _tentacle_arch_operation(root_folder, files_arch, branch_func, leaf_func):
        sub_dir_to_handle_coroutines = []
        for root, modules in files_arch.items():
            current_root = join(root_folder, root)
            await branch_func(current_root, modules)
            if isinstance(modules, dict):
                sub_dir_to_handle_coroutines.append(
                    TentaclesSetupManager._tentacle_arch_operation(current_root, modules, branch_func, leaf_func))
            elif leaf_func is not None:
                sub_dir_to_handle_coroutines += [leaf_func(join(current_root, directory))
                                                 for directory in modules]
        await gather(*sub_dir_to_handle_coroutines)

    @staticmethod
    async def _create_missing_files_and_folders(current_root, modules):
        await find_or_create(current_root)
        # create python init file
        await find_or_create_module_init_file(current_root, modules)

    @staticmethod
    async def _remove_tentacles_arch_init_file(current_root, _=None):
        potential_init_file = join(current_root, PYTHON_INIT_FILE)
        if exists(potential_init_file):
            remove(potential_init_file)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import aiohttp
import pytest
from os import path

from octobot_tentacles_manager.constants import TENTACLES_PATH
from octobot_tentacles_manager.api.installer import install_tentacles
from octobot_tentacles_manager.api.updater import update_all_tentacles, update_tentacles
from octobot_tentacles_manager.managers.tentacles_setup_manager import TentaclesSetupManager
from octobot_tentacles_manager.models.tentacle_factory import TentacleFactory

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_update_all_tentacles():
    async with aiohttp.ClientSession() as session:
        assert await update_all_tentacles(_tentacles_local_path(), aiohttp_session=session) == 0
    _cleanup()


async def test_update_one_tentacle_with_requirement():
    async with aiohttp.ClientSession() as session:
        assert await install_tentacles(["reddit_service_feed"], _tentacles_local_path(),
                                       aiohttp_session=session) == 0
        factory = TentacleFactory(TENTACLES_PATH)
        assert path.exists(path.join("tentacles", "Services", "Services_bases", "reddit_service", "reddit_service.py"))
        import tentacles.Services.Services_bases.reddit_service as red
        import tentacles.Services.Services_feeds.reddit_service_feed as feed
        dtm_tentacle_data = await factory.create_and_load_tentacle_from_module(feed)
        assert dtm_tentacle_data.version == "1.2.0"
        assert await update_tentacles(["reddit_service_feed"], _tentacles_update_local_path(),
                                      aiohttp_session=session) == 0
        dtm_tentacle_data = await factory.create_and_load_tentacle_from_module(feed)
        assert dtm_tentacle_data.version == "1.3.0"
    # _cleanup()


def _tentacles_local_path():
    return path.join("tests", "static", "tentacles.zip")


def _tentacles_update_local_path():
    return path.join("tests", "static", "update_tentacles.zip")


def _cleanup():
    TentaclesSetupManager.delete_tentacles_arch()



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import aiohttp
import pytest
from logging import INFO
from shutil import rmtree
from os import walk, path

from octobot_commons.logging.logging_util import set_logging_level
from octobot_tentacles_manager.constants import USER_TENTACLE_CONFIG_PATH, USER_TENTACLE_SPECIFIC_CONFIG_PATH, \
    TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR, USER_TENTACLE_CONFIG_FILE_PATH, TENTACLES_PATH, DEFAULT_BOT_PATH
from octobot_tentacles_manager.workers.install_worker import InstallWorker

# All test coroutines will be treated as marked.
from octobot_tentacles_manager.models.tentacle import Tentacle
from octobot_tentacles_manager.util.tentacle_fetching import fetch_and_extract_tentacles

pytestmark = pytest.mark.asyncio

temp_dir = "temp_tests"


async def test_install_two_tentacles():
    _cleanup()
    _enable_loggers()
    await fetch_and_extract_tentacles(temp_dir, path.join("tests", "static", "tentacles.zip"), None)
    worker = InstallWorker(temp_dir, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    worker.tentacles_setup_manager.default_tentacle_config = \
        path.join("tests", "static", "default_tentacle_config.json")
    assert await worker.process(["instant_fluctuations_evaluator", "generic_exchange_importer"]) == 0

    # test installed files
    trading_mode_files_count = sum(1 for _ in walk(path.join(TENTACLES_PATH, "Trading", "Mode")))
    assert trading_mode_files_count == 1
    backtesting_mode_files_count = sum(1 for _ in walk(path.join(TENTACLES_PATH, "Backtesting", "importers")))
    assert backtesting_mode_files_count == 7
    config_files = [f for f in walk(USER_TENTACLE_SPECIFIC_CONFIG_PATH)]
    config_files_count = len(config_files)
    assert config_files_count == 1
    assert "InstantFluctuationsEvaluator.json" in config_files[0][2]
    assert "DailyTradingMode.json" not in config_files[0][2]
    assert len(config_files[0][2]) == 1

    # test tentacles config
    with open(USER_TENTACLE_CONFIG_FILE_PATH, "r") as config_f:
        assert json.load(config_f) == {
            'tentacle_activation': {
                'InstantFluctuationsEvaluator': True
            }
        }
    _cleanup()


async def test_install_one_tentacle_with_requirement():
    async with aiohttp.ClientSession() as session:
        _cleanup()
        _enable_loggers()
        await fetch_and_extract_tentacles(temp_dir, path.join("tests", "static", "tentacles.zip"), None)
        worker = InstallWorker(temp_dir, TENTACLES_PATH, DEFAULT_BOT_PATH, False, session)
        worker.tentacles_setup_manager.default_tentacle_config = \
            path.join("tests", "static", "default_tentacle_config.json")
        assert await worker.process(["reddit_service_feed"]) == 0

    # test removed temporary requirements files
    assert not path.exists(TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR)

    # test installed files
    trading_mode_files_count = sum(1 for _ in walk(path.join(TENTACLES_PATH, "Trading", "Mode")))
    assert trading_mode_files_count == 1
    config_files = [f for f in walk(USER_TENTACLE_SPECIFIC_CONFIG_PATH)]
    assert len(config_files) == 1
    assert len(config_files[0][2]) == 0

    # test tentacles config
    with open(USER_TENTACLE_CONFIG_FILE_PATH, "r") as config_f:
        assert json.load(config_f) == {
            'tentacle_activation': {}
        }
    assert path.exists(path.join("tentacles", "Services", "Services_bases", "reddit_service", "reddit_service.py"))
    _cleanup()


async def test_install_all_tentacles():
    _cleanup()
    _enable_loggers()
    await fetch_and_extract_tentacles(temp_dir, path.join("tests", "static", "tentacles.zip"), None)
    worker = InstallWorker(temp_dir, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    worker.tentacles_setup_manager.default_tentacle_config = \
        path.join("tests", "static", "default_tentacle_config.json")
    assert await worker.process() == 0

    # test installed files
    trading_mode_files_count = sum(1 for _ in walk(path.join(TENTACLES_PATH, "Trading", "Mode")))
    assert trading_mode_files_count == 5
    config_files = [f for f in walk(USER_TENTACLE_SPECIFIC_CONFIG_PATH)]
    config_files_count = len(config_files)
    assert config_files_count == 1
    assert "DailyTradingMode.json" in config_files[0][2]
    assert len(config_files[0][2]) == 5

    # test tentacles config
    with open(USER_TENTACLE_CONFIG_FILE_PATH, "r") as config_f:
        assert json.load(config_f) == {
            'tentacle_activation': {
                'InstantFluctuationsEvaluator': True,
                'OtherInstantFluctuationsEvaluator': False,
                'SecondOtherInstantFluctuationsEvaluator': False,
                'RedditForumEvaluator': False,
                'SimpleMixedStrategyEvaluator': True,
                'DailyTradingMode': True
            }
        }
    _cleanup()


async def test_install_all_tentacles_twice():
    _cleanup()
    await fetch_and_extract_tentacles(temp_dir, path.join("tests", "static", "tentacles.zip"), None)
    worker = InstallWorker(temp_dir, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    worker.tentacles_setup_manager.default_tentacle_config = \
        path.join("tests", "static", "default_tentacle_config.json")
    assert await worker.process() == 0
    assert await worker.process() == 0
    trading_mode_files_count = sum(1 for _ in walk(path.join(TENTACLES_PATH, "Trading", "Mode")))
    assert trading_mode_files_count == 5
    _cleanup()


async def test_install_all_tentacles_fetching_requirements():
    async with aiohttp.ClientSession() as session:
        _cleanup()
        _enable_loggers()
        await fetch_and_extract_tentacles(temp_dir, path.join("tests", "static", "requirements_tentacles.zip"), None)
        worker = InstallWorker(temp_dir, TENTACLES_PATH, DEFAULT_BOT_PATH, False, session)
        worker.tentacles_setup_manager.default_tentacle_config = \
            path.join("tests", "static", "default_tentacle_config.json")
        assert await worker.process() == 0

    trading_mode_files_count = sum(1 for _ in walk(path.join(TENTACLES_PATH, "Trading", "Mode")))
    assert trading_mode_files_count == 5
    config_files = [f for f in walk(USER_TENTACLE_SPECIFIC_CONFIG_PATH)]
    config_files_count = len(config_files)
    assert config_files_count == 1
    # ensure fetched InstantFluctuationsEvaluator requirement
    assert "InstantFluctuationsEvaluator.json" in config_files[0][2]
    assert path.exists(path.join("tentacles", "Evaluator", "RealTime",
                                 "instant_fluctuations_evaluator", "instant_fluctuations_evaluator.py"))
    assert len(config_files[0][2]) == 4
    _cleanup()


def _enable_loggers():
    set_logging_level([clazz.__name__ for clazz in [InstallWorker, Tentacle]], INFO)


def _cleanup():
    if path.exists(temp_dir):
        rmtree(temp_dir)
    if path.exists(TENTACLES_PATH):
        rmtree(TENTACLES_PATH)
    if path.exists(USER_TENTACLE_CONFIG_PATH):
        rmtree(USER_TENTACLE_CONFIG_PATH)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from os.path import sep


class TentacleType:

    def __init__(self, type_path, module_name=None):
        self.path = type_path
        self.module_name = module_name

    @staticmethod
    def from_import_path(root, import_path):
        module_name = import_path.split(".")[-1]
        tentacle_type_path = import_path.split(f"{root}.")[-1]\
            .replace(f".{module_name}", "")\
            .replace(f".", sep)
        return TentacleType(tentacle_type_path, module_name)

    def get_last_element(self):
        return self.path.split(sep)[-1]

    def to_path(self):
        return self.path

    def __str__(self):
        return self.path.replace(sep, ".")



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from os.path import join
from octobot_commons.constants import USER_FOLDER, CONFIG_FOLDER

# Default tests tentacles URL
DEFAULT_TENTACLES_URL = "https://www.tentacles.octobot.online/repository/tentacles/officials/base/tests/0.0.2-tests.zip"

# Tentacles files
PYTHON_INIT_FILE = "__init__.py"
PYTHON_EXT = ".py"
CONFIG_EVALUATOR_FILE = "evaluator_config.json"
CONFIG_TRADING_FILE = "trading_config.json"
CONFIG_TENTACLES_FILE = "tentacles_config.json"
CONFIG_EXT = ".json"
DOCUMENTATION_EXT = ".md"
CONFIG_SCHEMA_EXT = "_schema.json"
TENTACLE_METADATA = "metadata.json"
DEFAULT_TENTACLE_CONFIG = join(CONFIG_FOLDER, "default_tentacles_config.json")
# tentacles setup folders back list
FOLDERS_BLACK_LIST = ["__pycache__"]

# Metadata keys
METADATA_VERSION = "version"
METADATA_TENTACLES = "tentacles"
METADATA_ORIGIN_PACKAGE = "origin_package"
METADATA_DEV_MODE = "dev_mode"
METADATA_TENTACLES_REQUIREMENTS = "tentacles-requirements"

# Requirements
TENTACLE_REQUIREMENT_VERSION_EQUALS = "=="

# Tentacle user config files and folders
USER_TENTACLE_CONFIG_PATH = join(USER_FOLDER, "tentacles_config")
USER_TENTACLE_CONFIG_FILE_PATH = join(USER_TENTACLE_CONFIG_PATH, CONFIG_TENTACLES_FILE)
USER_TENTACLE_SPECIFIC_CONFIG_PATH = join(USER_TENTACLE_CONFIG_PATH, "specific_config")

# Current minimum default tentacles version
TENTACLE_CURRENT_MINIMUM_DEFAULT_TENTACLES_VERSION = "1.2.0"
DEFAULT_TENTACLES_PACKAGE = "OctoBot-Default-Tentacles"

# Tentacles installation folders
TENTACLES_INSTALL_TEMP_DIR = "temp_tentacles"
TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR = "requirements_temp_tentacles"
TENTACLES_ARCHIVE_ROOT = "reference_tentacles"

# Tentacles folders
TENTACLES_PATH = "tentacles"
DEFAULT_BOT_PATH = "."
TENTACLES_BACKTESTING_PATH = "Backtesting"
TENTACLES_EVALUATOR_PATH = "Evaluator"
TENTACLES_INTERFACES_PATH = "Interfaces"
TENTACLES_NOTIFIERS_PATH = "Notifiers"
TENTACLES_SERVICES_PATH = "Services"
TENTACLES_SERVICES_BASES_PATH = "Services_bases"
TENTACLES_SERVICES_FEEDS_PATH = "Services_feeds"
TENTACLES_TRADING_PATH = "Trading"

# Tentacles sub-folders
TENTACLE_MAX_SUB_FOLDERS_LEVEL = 3
TENTACLES_BACKTESTING_COLLECTORS_PATH = "collectors"
TENTACLES_BACKTESTING_CONVERTERS_PATH = "converters"
TENTACLES_BACKTESTING_IMPORTERS_PATH = "importers"
TENTACLES_BACKTESTING_THIRD_LEVEL_EXCHANGES_PATH = "exchanges"
TENTACLES_BACKTESTING_THIRD_LEVEL_SOCIAL_PATH = "social"
TENTACLES_EVALUATOR_REALTIME_PATH = "RealTime"
TENTACLES_EVALUATOR_TA_PATH = "TA"
TENTACLES_EVALUATOR_SOCIAL_PATH = "Social"
TENTACLES_EVALUATOR_STRATEGIES_PATH = "Strategies"
TENTACLES_EVALUATOR_UTIL_PATH = "Util"
TENTACLES_TRADING_MODE_PATH = "Mode"
TENTACLES_TRADING_EXCHANGE_PATH = "Exchange"
TENTACLES_WEBSOCKETS_FEEDS_PATH = "feeds"

# Tentacle local module folders
TENTACLE_CONFIG = "config"
TENTACLE_RESOURCES = "resources"
TENTACLE_TESTS = "tests"

# Tentacle creator
TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER = "creator_tentacles_temp"
TENTACLES_PACKAGE_FORMAT = "zip"
PYTHON_GENERATED_ELEMENTS = ["__pycache__"]
PYTHON_GENERATED_ELEMENTS_EXTENSION = ["pyc"]
TENTACLE_TEMPLATE_PATH = "templates"
TENTACLE_TEMPLATE_DESCRIPTION = "description"
TENTACLE_CONFIG_TEMPLATE_PRE_EXT = "_config"
TENTACLE_TEMPLATE_EXT = ".template"
TENTACLE_TEMPLATE_PRE_EXT = "_tentacle"

# Tentacles that can be activated / deactivated
ACTIVATABLE_TENTACLES = [
    TENTACLES_EVALUATOR_REALTIME_PATH,
    TENTACLES_EVALUATOR_SOCIAL_PATH,
    TENTACLES_EVALUATOR_TA_PATH,
    TENTACLES_EVALUATOR_STRATEGIES_PATH,
    TENTACLES_TRADING_MODE_PATH,
]

# Tentacles architecture
TENTACLES_FOLDERS_ARCH = {
    TENTACLES_BACKTESTING_PATH: {
        TENTACLES_BACKTESTING_COLLECTORS_PATH: [
            TENTACLES_BACKTESTING_THIRD_LEVEL_EXCHANGES_PATH,
            TENTACLES_BACKTESTING_THIRD_LEVEL_SOCIAL_PATH
        ],
        TENTACLES_BACKTESTING_CONVERTERS_PATH: [
            TENTACLES_BACKTESTING_THIRD_LEVEL_EXCHANGES_PATH
        ],
        TENTACLES_BACKTESTING_IMPORTERS_PATH: [
            TENTACLES_BACKTESTING_THIRD_LEVEL_EXCHANGES_PATH,
            TENTACLES_BACKTESTING_THIRD_LEVEL_SOCIAL_PATH
        ]
    },
    TENTACLES_EVALUATOR_PATH: [
        TENTACLES_EVALUATOR_REALTIME_PATH,
        TENTACLES_EVALUATOR_SOCIAL_PATH,
        TENTACLES_EVALUATOR_TA_PATH,
        TENTACLES_EVALUATOR_STRATEGIES_PATH,
        TENTACLES_EVALUATOR_UTIL_PATH
    ],
    TENTACLES_SERVICES_PATH: [
        TENTACLES_INTERFACES_PATH,
        TENTACLES_NOTIFIERS_PATH,
        TENTACLES_SERVICES_BASES_PATH,
        TENTACLES_SERVICES_FEEDS_PATH
    ],
    TENTACLES_TRADING_PATH: [
        TENTACLES_TRADING_MODE_PATH,
        TENTACLES_TRADING_EXCHANGE_PATH
    ]
}

TENTACLE_MODULE_FOLDERS = {
    TENTACLE_CONFIG,
    TENTACLE_RESOURCES,
    TENTACLE_TESTS
}


# tentacles files management
TENTACLE_TYPES = [
    TENTACLES_BACKTESTING_PATH,
    TENTACLES_EVALUATOR_PATH,
    TENTACLES_SERVICES_PATH,
    TENTACLES_TRADING_PATH
]



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
from os.path import join

import aiofiles

from octobot_tentacles_manager.constants import TENTACLE_METADATA, METADATA_VERSION, METADATA_ORIGIN_PACKAGE, \
    METADATA_TENTACLES, METADATA_TENTACLES_REQUIREMENTS, TENTACLE_REQUIREMENT_VERSION_EQUALS, METADATA_DEV_MODE


class Tentacle:
    def __init__(self, tentacle_root_path, name, tentacle_type):
        self.tentacle_root_path = tentacle_root_path
        self.name = name
        self.tentacle_type = tentacle_type
        self.tentacle_path = join(self.tentacle_root_path, self.tentacle_type.to_path())
        self.version = None
        self.tentacle_class_names = None
        self.origin_package = None
        self.tentacles_requirements = None
        self.in_dev_mode = False
        self.metadata = {}

    async def initialize(self):
        async with aiofiles.open(join(self.tentacle_path, self.name, TENTACLE_METADATA), "r") as metadata_file:
            self.metadata = json.loads(await metadata_file.read())
            self.version = self.metadata[METADATA_VERSION]
            self.origin_package = self.metadata[METADATA_ORIGIN_PACKAGE]
            self.tentacle_class_names = self.metadata[METADATA_TENTACLES]
            self.tentacles_requirements = self.metadata[METADATA_TENTACLES_REQUIREMENTS]
            if METADATA_DEV_MODE in self.metadata:
                self.in_dev_mode = self.metadata[METADATA_DEV_MODE]

    @staticmethod
    def find(iterable, name):
        for tentacle in iterable:
            if tentacle.name == name:
                return tentacle
        return None

    def is_valid(self):
        return self.version is not None

    def get_simple_tentacle_type(self):
        return self.tentacle_type.get_last_element()

    def __str__(self):
        str_rep = f"{self.name} tentacle [type: {self.tentacle_type}"
        if self.is_valid():
            return f"{str_rep}, version: {self.version}]"
        else:
            return f"{str_rep}]"

    def extract_tentacle_requirements(self):
        return [self._parse_requirements(component) for component in self.tentacles_requirements]

    @staticmethod
    def _parse_requirements(requirement):
        if TENTACLE_REQUIREMENT_VERSION_EQUALS in requirement:
            return requirement.split(TENTACLE_REQUIREMENT_VERSION_EQUALS)
        else:
            return [requirement, None]



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from asyncio import gather

from octobot_tentacles_manager.managers.tentacle_manager import TentacleManager
from octobot_tentacles_manager.managers.tentacles_init_files_manager import update_tentacle_type_init_file
from octobot_tentacles_manager.workers.tentacles_worker import TentaclesWorker
from octobot_tentacles_manager.util.tentacle_explorer import load_tentacle_with_metadata


class UninstallWorker(TentaclesWorker):

    async def process(self, name_filter=None) -> int:
        self.reset_worker()
        if self.confirm_action("Remove all installed tentacles ?"
                               if name_filter is None else "Remove {', '.join(name_filter)} tentacles ?"):
            if name_filter is None:
                self.tentacles_setup_manager.delete_tentacles_arch(force=True, with_user_config=False,
                                                                   bot_installation_path=self.bot_installation_path)
            else:
                self.progress = 1
                all_tentacles = await load_tentacle_with_metadata(self.tentacle_path)
                self.register_error_on_missing_tentacles(all_tentacles, name_filter)
                to_uninstall_tentacles = [tentacle
                                          for tentacle in all_tentacles
                                          if tentacle.name in name_filter]
                await gather(*[self._uninstall_tentacle(tentacle) for tentacle in to_uninstall_tentacles])
            await self.tentacles_setup_manager.create_missing_tentacles_arch()
            await self.tentacles_setup_manager.refresh_user_tentacles_setup_config_file()
            self.log_summary()
        return len(self.errors)

    async def _uninstall_tentacle(self, tentacle):
        try:
            tentacle_manager = TentacleManager(tentacle)
            await tentacle_manager.uninstall_tentacle()
            update_tentacle_type_init_file(tentacle, tentacle.tentacle_path, remove_import=True)
            self.logger.info(f"[{self.progress}/{self.total_steps}] uninstalled {tentacle}")
        except Exception as e:
            message = f"Error when uninstalling {tentacle.name}: {e}"
            self.errors.append(message)
            self.logger.exception(e, True, message)
        finally:
            self.progress += 1



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import pytest
from logging import INFO
from shutil import rmtree
from os import walk, path

from octobot_commons.logging.logging_util import set_logging_level
from octobot_tentacles_manager.constants import USER_TENTACLE_CONFIG_PATH, TENTACLES_PATH, \
    USER_TENTACLE_CONFIG_FILE_PATH, DEFAULT_BOT_PATH
from octobot_tentacles_manager.workers.install_worker import InstallWorker

# All test coroutines will be treated as marked.
from octobot_tentacles_manager.models.tentacle import Tentacle
from octobot_tentacles_manager.workers.uninstall_worker import UninstallWorker
from octobot_tentacles_manager.util.tentacle_fetching import fetch_and_extract_tentacles

pytestmark = pytest.mark.asyncio

temp_dir = "temp_tests"


async def test_uninstall_two_tentacles():
    _cleanup()
    _enable_loggers()
    await fetch_and_extract_tentacles(temp_dir, path.join("tests", "static", "tentacles.zip"), None)
    install_worker = InstallWorker(temp_dir, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    install_worker.tentacles_setup_manager.default_tentacle_config \
        = path.join("tests", "static", "default_tentacle_config.json")
    assert await install_worker.process() == 0
    tentacles_files_count = sum(1 for _ in walk(TENTACLES_PATH))
    assert tentacles_files_count > 60

    uninstall_worker = UninstallWorker(None, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    uninstall_worker.tentacles_setup_manager.default_tentacle_config = \
        path.join("tests", "static", "default_tentacle_config.json")
    assert await uninstall_worker.process(["instant_fluctuations_evaluator", "generic_exchange_importer"]) == 0
    tentacles_files_count = sum(1 for _ in walk(TENTACLES_PATH))
    assert tentacles_files_count < 60
    with open(USER_TENTACLE_CONFIG_FILE_PATH, "r") as config_f:
        assert json.load(config_f) == {
            'tentacle_activation': {
                'DailyTradingMode': True,
                'OtherInstantFluctuationsEvaluator': False,
                'SecondOtherInstantFluctuationsEvaluator': False,
                'RedditForumEvaluator': False,
                'SimpleMixedStrategyEvaluator': True
            }
        }
    _cleanup()


async def test_uninstall_all_tentacles():
    _cleanup()
    _enable_loggers()
    await fetch_and_extract_tentacles(temp_dir, path.join("tests", "static", "tentacles.zip"), None)
    install_worker = InstallWorker(temp_dir, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    install_worker.tentacles_setup_manager.default_tentacle_config = \
        path.join("tests", "static", "default_tentacle_config.json")
    assert await install_worker.process() == 0
    tentacles_files_count = sum(1 for _ in walk(TENTACLES_PATH))
    assert tentacles_files_count > 60

    uninstall_worker = UninstallWorker(None, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    uninstall_worker.tentacles_setup_manager.default_tentacle_config = \
        path.join("tests", "static", "default_tentacle_config.json")
    assert await uninstall_worker.process() == 0
    tentacles_files_count = sum(1 for _ in walk(TENTACLES_PATH))
    assert tentacles_files_count == 24
    with open(USER_TENTACLE_CONFIG_FILE_PATH, "r") as config_f:
        assert json.load(config_f) == {
            'tentacle_activation': {}
        }
    _cleanup()


def _enable_loggers():
    set_logging_level([clazz.__name__ for clazz in [InstallWorker, Tentacle]], INFO)


def _cleanup():
    if path.exists(temp_dir):
        rmtree(temp_dir)
    if path.exists(TENTACLES_PATH):
        rmtree(TENTACLES_PATH)
    if path.exists(USER_TENTACLE_CONFIG_PATH):
        rmtree(USER_TENTACLE_CONFIG_PATH)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
from os.path import exists
from shutil import rmtree
import aiohttp
import pytest
from os import path, remove, mkdir

from octobot_tentacles_manager.api.creator import create_tentacles_package
from octobot_tentacles_manager.api.installer import install_all_tentacles
from octobot_tentacles_manager.constants import TENTACLES_PATH, PYTHON_INIT_FILE, TENTACLES_EVALUATOR_PATH, \
    TENTACLES_EVALUATOR_REALTIME_PATH, TENTACLE_METADATA, METADATA_VERSION, METADATA_ORIGIN_PACKAGE, \
    METADATA_TENTACLES, METADATA_TENTACLES_REQUIREMENTS, METADATA_DEV_MODE, TENTACLES_TRADING_PATH, \
    TENTACLES_TRADING_MODE_PATH, TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER
from octobot_tentacles_manager.managers.tentacles_setup_manager import TentaclesSetupManager

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio
TENTACLE_PACKAGE = "tentacle_package"


async def test_create_folder_tentacles_package():
    _cleanup()
    async with aiohttp.ClientSession() as session:
        assert await install_all_tentacles(_tentacles_local_path(), aiohttp_session=session) == 0

    # set instant_fluctuations_evaluator in dev mode
    tentacle_path = path.join(TENTACLES_PATH, TENTACLES_EVALUATOR_PATH, TENTACLES_EVALUATOR_REALTIME_PATH)
    with open(path.join(tentacle_path, "instant_fluctuations_evaluator", TENTACLE_METADATA), "w+") as metadata:
        new_metadata = {
          METADATA_VERSION: "1.2.0",
          METADATA_ORIGIN_PACKAGE: "OctoBot-Default-Tentacles",
          METADATA_TENTACLES: ["InstantFluctuationsEvaluator"],
          METADATA_TENTACLES_REQUIREMENTS: [],
          METADATA_DEV_MODE: True
        }
        json.dump(new_metadata, metadata)

    # add generated python file
    random_content = "123"
    generated_file_path = path.join(TENTACLES_PATH, "file.pyc")
    with open(generated_file_path, "w+") as generated_file:
        generated_file.write(random_content)
    # add generated python folder
    generated_folder_path = path.join(TENTACLES_PATH, TENTACLES_TRADING_PATH, "__pycache__")
    mkdir(generated_folder_path)

    # create folder to force folder merge
    mkdir(TENTACLE_PACKAGE)
    mkdir(path.join(TENTACLE_PACKAGE, TENTACLES_TRADING_PATH))
    with open(path.join(TENTACLE_PACKAGE, "rnd"), "w+") as rand_file:
        rand_file.write(random_content)

    assert await create_tentacles_package(TENTACLE_PACKAGE, in_zip=False) == 0
    assert path.exists(TENTACLE_PACKAGE)

    # random init file still here
    with open(path.join(TENTACLE_PACKAGE, "rnd")) as rand_file:
        assert random_content == rand_file.read()

    # generated file not copied
    assert not path.exists(path.join(TENTACLE_PACKAGE, "file.pyc"))

    # generated folder not copied
    assert not path.exists(path.join(TENTACLE_PACKAGE, TENTACLES_TRADING_PATH, "__pycache__"))

    # did not add instant_fluctuations_evaluator (in dev-mode)
    package_tentacle_path = path.join(TENTACLE_PACKAGE, TENTACLES_EVALUATOR_PATH, TENTACLES_EVALUATOR_REALTIME_PATH)
    assert not path.exists(path.join(package_tentacle_path, "instant_fluctuations_evaluator"))

    # added other_instant_fluctuations_evaluator
    assert path.exists(path.join(package_tentacle_path, "other_instant_fluctuations_evaluator"))

    # added daily_trading_mode
    trading_mode_tentacle_path = path.join(TENTACLE_PACKAGE, TENTACLES_TRADING_PATH, TENTACLES_TRADING_MODE_PATH)
    assert path.exists(path.join(trading_mode_tentacle_path, "daily_trading_mode"))
    assert path.exists(path.join(trading_mode_tentacle_path, "daily_trading_mode", "daily_trading_mode.py"))

    # removed python init files
    assert not path.exists(path.join(TENTACLE_PACKAGE, PYTHON_INIT_FILE))
    evaluator_path = path.join(TENTACLE_PACKAGE, TENTACLES_EVALUATOR_PATH)
    assert not path.exists(path.join(evaluator_path, PYTHON_INIT_FILE))
    assert not path.exists(path.join(evaluator_path, TENTACLES_EVALUATOR_REALTIME_PATH, PYTHON_INIT_FILE))

    _cleanup()


async def test_create_zipped_tentacles_package():
    _cleanup()
    tentacle_package = "tentacle_package.zip"
    async with aiohttp.ClientSession() as session:
        assert await install_all_tentacles(_tentacles_local_path(), aiohttp_session=session) == 0
    assert await create_tentacles_package(tentacle_package) == 0
    assert path.exists(tentacle_package)
    assert not path.exists(TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER)
    remove(tentacle_package)
    _cleanup()


def _tentacles_local_path():
    return path.join("tests", "static", "tentacles.zip")


def _cleanup():
    if exists(TENTACLES_PATH):
        TentaclesSetupManager.delete_tentacles_arch(force=True)
    if exists(TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER):
        rmtree(TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER)
    if exists(TENTACLE_PACKAGE):
        rmtree(TENTACLE_PACKAGE)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import argparse
import asyncio
import sys
import aiohttp

from octobot_commons.logging.logging_util import get_logger
from octobot_tentacles_manager.api.creator import start_tentacle_creator, create_tentacles_package
from octobot_tentacles_manager.api.installer import install_all_tentacles, install_tentacles, \
    USER_HELP as INSTALL_USER_HELP, install_single_tentacle, repair_installation
from octobot_tentacles_manager.api.uninstaller import uninstall_all_tentacles, uninstall_tentacles, \
    USER_HELP as UNINSTALL_USER_HELP
from octobot_tentacles_manager.api.updater import update_all_tentacles, update_tentacles, \
    USER_HELP as UPDATE_USER_HELP
from octobot_tentacles_manager.constants import DEFAULT_TENTACLES_URL, DEFAULT_BOT_PATH
from octobot_tentacles_manager import PROJECT_NAME


async def _handle_package_manager_command(starting_args,
                                          tentacles_url,
                                          target_dir,
                                          single_tentacle_path,
                                          single_tentacle_type,
                                          export_tentacles_output,
                                          packed_tentacles_output) -> int:
    error_count = 0
    LOGGER = get_logger(f"{PROJECT_NAME}-CLI")
    async with aiohttp.ClientSession() as aiohttp_session:
        include_dev_mode = starting_args.include_dev_mode
        if starting_args.creator:
            error_count = start_tentacle_creator({}, starting_args.creator)
        elif starting_args.repair:
            error_count = await repair_installation(bot_path=target_dir)
        elif starting_args.export:
            error_count = await create_tentacles_package(export_tentacles_output, target_dir,
                                                         in_zip=False, with_dev_mode=include_dev_mode)
        elif starting_args.pack:
            error_count = await create_tentacles_package(packed_tentacles_output, target_dir,
                                                         in_zip=True, with_dev_mode=include_dev_mode)
        elif single_tentacle_path:
            error_count = await install_single_tentacle(single_tentacle_path,
                                                        single_tentacle_type,
                                                        bot_path=target_dir,
                                                        aiohttp_session=aiohttp_session)
        elif not (starting_args.all or starting_args.tentacle_names):
            LOGGER.error("Please provide at least one tentacle name or add the '--all' parameter")
            return 1
        elif starting_args.install:
            if tentacles_url is None:
                LOGGER.error("Please provide a tentacle path or URL")
                return 1
            if starting_args.all:
                error_count = await install_all_tentacles(tentacles_url,
                                                          bot_path=target_dir,
                                                          aiohttp_session=aiohttp_session)
            else:
                error_count = await install_tentacles(starting_args.tentacle_names,
                                                      tentacles_url,
                                                      bot_path=target_dir,
                                                      aiohttp_session=aiohttp_session)
        elif starting_args.update:
            if tentacles_url is None:
                LOGGER.error("Please provide a tentacle path or URL")
                return 1
            if starting_args.all:
                error_count = await update_all_tentacles(tentacles_url,
                                                         bot_path=target_dir,
                                                         aiohttp_session=aiohttp_session)
            else:
                error_count = await update_tentacles(starting_args.tentacle_names,
                                                     tentacles_url,
                                                     bot_path=target_dir,
                                                     aiohttp_session=aiohttp_session)
        elif starting_args.uninstall:
            if starting_args.all:
                error_count = await uninstall_all_tentacles(bot_path=target_dir,
                                                            use_confirm_prompt=starting_args.force)
            else:
                error_count = await uninstall_tentacles(starting_args.tentacle_names,
                                                        bot_path=target_dir,
                                                        use_confirm_prompt=starting_args.force)
    if error_count > 0:
        LOGGER.error(f"{error_count} errors occurred while processing tentacles.")
        return 1
    return 0


def handle_tentacles_manager_command(starting_args,
                                     tentacles_url=None,
                                     single_tentacle_path=None,
                                     single_tentacle_type=None,
                                     export_tentacles_output=None,
                                     packed_tentacles_output=None) -> int:
    tentacles_url = starting_args.location[0] if starting_args.location else tentacles_url
    target_bot_dir = starting_args.directory[0] if starting_args.directory else DEFAULT_BOT_PATH
    if starting_args.single_tentacle_install:
        single_tentacle_path = starting_args.single_tentacle_install[0]
        single_tentacle_type = starting_args.single_tentacle_install[1]
    export_tentacles_output = starting_args.export[0] if starting_args.export else export_tentacles_output
    packed_tentacles_output = starting_args.pack[0] if starting_args.pack else packed_tentacles_output
    return asyncio.run(_handle_package_manager_command(starting_args,
                                                       tentacles_url,
                                                       target_bot_dir,
                                                       single_tentacle_path,
                                                       single_tentacle_type,
                                                       export_tentacles_output,
                                                       packed_tentacles_output))


def register_tentacles_manager_arguments(tentacles_parser) -> None:
    tentacles_parser.add_argument("-i", "--install", help=INSTALL_USER_HELP, action='store_true')
    tentacles_parser.add_argument("-u", "--update", help=UPDATE_USER_HELP, action='store_true')
    tentacles_parser.add_argument("-ui", "--uninstall", help=UNINSTALL_USER_HELP, action='store_true')
    tentacles_parser.add_argument("-r", "--repair", help="Repair a tentacles installation. "
                                                         "Fixes __init__.py files, missing main folders "
                                                         "and configuration files.", action='store_true')
    tentacles_parser.add_argument("-sti", "--single-tentacle-install", help='Single tentacle to install identified by '
                                                                            'a local path and its tentacle type.\n'
                                                                            'Example: -sti "/bot/macd_eval" '
                                                                            '"Evaluator/TA".',
                                  nargs=2)
    tentacles_parser.add_argument("-p", "--pack", help="Create a tentacle package containing all tentacles "
                                                       "in the given folder.\n"
                                                       "Specify source folder using the --directory argument.\n"
                                                       "Example: -p myTentaclesPackage.zip -d ./tentacles", nargs=1)
    tentacles_parser.add_argument("-e", "--export", help="Export tentacles into a folder containing all tentacles "
                                                         "in the given folder. Removes installation generated files. "
                                                         "Specify source folder using the --directory argument.\n"
                                                         "Example: -e myTentaclesFolder -d ./tentacles", nargs=1)
    tentacles_parser.add_argument("-a", "--all", help="Apply command to all available Tentacles.", action='store_true')
    tentacles_parser.add_argument("-d", "--directory", help=f"Path to the root of the OctoBot installation folder "
                                                            f"to operate on. Default is '{DEFAULT_BOT_PATH}'.",
                                  nargs=1)
    tentacles_parser.add_argument("-l", "--location", help="Tentacles package local path or url to find the "
                                                           "tentacles package to process.", nargs=1)
    tentacles_parser.add_argument("-f", "--force", help="Skip user confirmations.", action='store_true')
    tentacles_parser.add_argument("-idm", "--include-dev-mode", help="Include tentacles in dev mode in export and "
                                                                     "pack commands.", action='store_true')
    tentacles_parser.add_argument("-c", "--creator", help="Start OctoBot Tentacles Creator.\n Examples: -c Evaluator "
                                                          "to create a new evaluator tentacles. Use: -c help to get the"
                                                          " Tentacle Creator help.", nargs='+')
    tentacles_parser.add_argument("tentacle_names", nargs="*")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=f"{PROJECT_NAME}-CLI")
    register_tentacles_manager_arguments(parser)
    args = parser.parse_args(sys.argv[1:])
    sys.exit(handle_tentacles_manager_command(args))



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from os.path import join
from asyncio import gather, Event, wait_for

from octobot_commons.logging.logging_util import get_logger
from octobot_tentacles_manager.managers.tentacle_manager import TentacleManager
from octobot_tentacles_manager.managers.tentacles_setup_manager import TentaclesSetupManager
from octobot_tentacles_manager.constants import TENTACLES_ARCHIVE_ROOT, DEFAULT_TENTACLES_URL
from octobot_tentacles_manager.util.tentacle_explorer import load_tentacle_with_metadata
from octobot_tentacles_manager.util.tentacle_fetching import fetch_and_extract_tentacles


class TentaclesWorker:
    TENTACLES_FETCHING_TIMEOUT = 120

    def __init__(self,
                 reference_tentacles_dir,
                 tentacle_path,
                 bot_installation_path,
                 use_confirm_prompt,
                 aiohttp_session):
        self.logger = get_logger(self.__class__.__name__)
        self.aiohttp_session = aiohttp_session
        self.use_confirm_prompt = use_confirm_prompt

        self.reference_tentacles_root = join(reference_tentacles_dir, TENTACLES_ARCHIVE_ROOT) \
            if reference_tentacles_dir is not None else TENTACLES_ARCHIVE_ROOT
        self.bot_installation_path = bot_installation_path
        self.tentacle_path = join(bot_installation_path, tentacle_path)
        self.tentacles_setup_manager = TentaclesSetupManager(self.tentacle_path, self.bot_installation_path)
        self.tentacles_path_or_url = None

        self.total_steps = 0
        self.progress = 0
        self.errors = []

        self.to_process_tentacle_modules = {}
        self.processed_tentacles_modules = []

        self.fetched_for_requirements_tentacles = []
        self.fetched_for_requirements_tentacles_versions = {}
        self.fetching_requirements = False
        self.requirements_downloading_event = Event()

    def reset_worker(self):
        self.errors = []

        self.to_process_tentacle_modules = {}
        self.processed_tentacles_modules = []

        self.fetched_for_requirements_tentacles = []
        self.fetched_for_requirements_tentacles_versions = {}
        self.fetching_requirements = False
        self.requirements_downloading_event = Event()

    def log_summary(self):
        if self.errors:
            self.logger.info(" *** Error summary: ***")
            for error in self.errors:
                self.logger.error(f"Error when handling tentacle: {error}")
        else:
            self.logger.info(" *** All tentacles have been successfully processed ***")

    def register_to_process_tentacles_modules(self, to_process_tentacle):
        self.to_process_tentacle_modules = self._get_version_by_tentacle(to_process_tentacle)

    def register_error_on_missing_tentacles(self, all_tentacles, name_filter):
        if name_filter is not None:
            all_tentacles_names = [tentacle.name for tentacle in all_tentacles]
            missing_tentacles = [tentacle_name
                                 for tentacle_name in name_filter
                                 if tentacle_name not in all_tentacles_names]
            if missing_tentacles:
                self.errors.append(f"Tentacles: {', '.join(missing_tentacles)} can't be found.")

    async def handle_requirements(self, tentacle, callback):
        if tentacle.tentacles_requirements:
            await self._handle_tentacles_requirements(tentacle, callback)
        # TODO: handle python requirements

    async def _handle_tentacles_requirements(self, tentacle, callback):
        missing_requirements = TentacleManager.find_tentacles_missing_requirements(tentacle,
                                                                                   self.to_process_tentacle_modules)
        if missing_requirements:
            if not self.fetching_requirements:
                self.fetching_requirements = True
                await self._fetch_all_available_tentacles()
            else:
                await wait_for(self.requirements_downloading_event.wait(), self.TENTACLES_FETCHING_TIMEOUT)
            await callback(tentacle, missing_requirements)

    def confirm_action(self, action):
        if not self.use_confirm_prompt:
            return True
        else:
            confirmations = ["yes", "y"]
            user_input = input(f"{action} Y/N").lower()
            if user_input in confirmations:
                return True
            else:
                self.logger.info("Action aborted.")
                return False

    async def _fetch_all_available_tentacles(self):
        # try getting it from available tentacles
        await gather(*[self._fetch_tentacles_for_requirement(repo)
                       for repo in TentaclesSetupManager.get_available_tentacles_repos()])
        self.requirements_downloading_event.set()

    async def _fetch_tentacles_for_requirement(self, repo):
        await fetch_and_extract_tentacles(repo, self.tentacles_path_or_url or DEFAULT_TENTACLES_URL,
                                          self.aiohttp_session, merge_dirs=True)
        requirements_tentacles_path = join(repo, TENTACLES_ARCHIVE_ROOT)
        self.fetched_for_requirements_tentacles = await load_tentacle_with_metadata(requirements_tentacles_path)
        self.fetched_for_requirements_tentacles_versions = \
            self._get_version_by_tentacle(self.fetched_for_requirements_tentacles)

    @staticmethod
    def _get_version_by_tentacle(all_tentacles):
        return {tentacle.name: tentacle.version
                for tentacle in all_tentacles}



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import aiofiles
from zipfile import ZipFile
from os import remove, path
from shutil import rmtree

from octobot_tentacles_manager.constants import TENTACLE_TYPES, TENTACLES_ARCHIVE_ROOT

DOWNLOADED_DATA_CHUNK_SIZE = 60000


async def fetch_and_extract_tentacles(tentacles_temp_dir, tentacles_path_or_url, aiohttp_session, merge_dirs=False):
    compressed_file = tentacles_path_or_url
    should_download = _is_url(tentacles_path_or_url)
    if should_download:
        if aiohttp_session is None:
            raise RuntimeError("Missing aiohttp_session argument")
        compressed_file = f"downloaded_{tentacles_temp_dir}"
        await _download_tentacles(compressed_file, tentacles_path_or_url, aiohttp_session)
    await _extract_tentacles(compressed_file, tentacles_temp_dir, should_download, merge_dirs)


def cleanup_temp_dirs(target_path):
    if path.exists(target_path):
        rmtree(target_path)


async def _download_tentacles(target_file, download_URL, aiohttp_session):
    async with aiohttp_session.get(download_URL) as resp:
        async with aiofiles.open(target_file, 'wb+') as downloaded_file:
            while True:
                chunk = await resp.content.read(DOWNLOADED_DATA_CHUNK_SIZE)
                if not chunk:
                    # resp.content.read returns an empty chunk when completed
                    break
                await downloaded_file.write(chunk)


async def _extract_tentacles(source_path, target_path, remove_source_file, merge_dirs):
    if path.exists(target_path) and path.isdir(target_path) and not merge_dirs:
        rmtree(target_path)
    with ZipFile(source_path) as zipped_tentacles:
        for archive_member in zipped_tentacles.namelist():
            if _is_tentacle_valid_tentacle_file(archive_member):
                zipped_tentacles.extract(archive_member, target_path)
    if remove_source_file:
        remove(source_path)


def _is_tentacle_valid_tentacle_file(archive_member):
    member_path = archive_member.split("/")
    return len(member_path) >= 2 \
        and member_path[0] == TENTACLES_ARCHIVE_ROOT \
        and member_path[1] in TENTACLE_TYPES


def _is_url(string):
    return string.startswith("https://") \
           or string.startswith("http://") \
           or string.startswith("ftp://") \
           or string.startswith("sftp://")



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from os.path import isdir, isfile

import aiohttp
import pytest
from importlib import reload
from os import path

from octobot_tentacles_manager.api.installer import install_all_tentacles
import octobot_tentacles_manager.loaders.tentacle_loading as tentacle_loading

# All test coroutines will be treated as marked.
from octobot_tentacles_manager.managers.tentacles_setup_manager import TentaclesSetupManager

pytestmark = pytest.mark.asyncio


async def test_get_tentacle_without_loading():
    # reload tentacle_loading module to force reset of cached tentacle data
    reload(tentacle_loading)
    async with aiohttp.ClientSession() as session:
        await install_all_tentacles(_tentacles_local_path(), aiohttp_session=session)
    # force tentacle data reset
    tentacle_loading._tentacle_by_tentacle_class = None
    with pytest.raises(RuntimeError):
        from tentacles.Services import RedditService
        tentacle_loading.get_tentacle(RedditService)
    _cleanup()


async def test_with_reload_tentacle_by_tentacle_class_installed_tentacles():
    async with aiohttp.ClientSession() as session:
        await install_all_tentacles(_tentacles_local_path(), aiohttp_session=session)
    await tentacle_loading.reload_tentacle_by_tentacle_class()
    from tentacles.Services import RedditService
    from tentacles.Trading.Mode import DailyTradingMode
    assert tentacle_loading.get_tentacle(RedditService) is not None
    assert tentacle_loading.get_tentacle(DailyTradingMode) is not None
    assert isdir(tentacle_loading.get_resources_path(DailyTradingMode))
    assert isdir(tentacle_loading.get_resources_path(RedditService))
    assert isfile(tentacle_loading.get_documentation_file_path(DailyTradingMode))
    _cleanup()


def _tentacles_local_path():
    return path.join("tests", "static", "tentacles.zip")


def _cleanup():
    TentaclesSetupManager.delete_tentacles_arch()



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import aiohttp
import pytest
from os import path, walk

from octobot_tentacles_manager.constants import TENTACLES_PATH
from octobot_tentacles_manager.api.installer import install_all_tentacles, install_tentacles
from octobot_tentacles_manager.api.uninstaller import uninstall_all_tentacles, uninstall_tentacles
from octobot_tentacles_manager.loaders.tentacle_loading import get_tentacle_classes
from octobot_tentacles_manager.managers.tentacles_setup_manager import TentaclesSetupManager

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_uninstall_all_tentacles():
    async with aiohttp.ClientSession() as session:
        assert await install_all_tentacles(_tentacles_local_path(), aiohttp_session=session) == 0
        trading_mode_files_count = sum(1 for _ in walk(TENTACLES_PATH))
        assert trading_mode_files_count > 50
    assert await uninstall_all_tentacles() == 0
    trading_mode_files_count = sum(1 for _ in walk(TENTACLES_PATH))
    assert trading_mode_files_count == 24
    _cleanup()


async def test_uninstall_one_tentacle():
    async with aiohttp.ClientSession() as session:
        assert await install_tentacles(["reddit_service"], _tentacles_local_path(), aiohttp_session=session) == 0
        trading_mode_files_count = sum(1 for _ in walk(TENTACLES_PATH))
        assert trading_mode_files_count > 25
    assert "RedditService" in get_tentacle_classes()
    assert await uninstall_tentacles(["reddit_service"]) == 0
    trading_mode_files_count = sum(1 for _ in walk(TENTACLES_PATH))
    assert trading_mode_files_count == 24
    assert "RedditService" not in get_tentacle_classes()
    _cleanup()


def _tentacles_local_path():
    return path.join("tests", "static", "tentacles.zip")


def _cleanup():
    TentaclesSetupManager.delete_tentacles_arch()



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from distutils.version import LooseVersion

from octobot_tentacles_manager.workers.install_worker import InstallWorker
from octobot_tentacles_manager.models.tentacle import Tentacle
from octobot_tentacles_manager.util.tentacle_explorer import load_tentacle_with_metadata


class UpdateWorker(InstallWorker):

    def __init__(self, reference_tentacles_dir, tentacle_path,
                 bot_installation_path, use_confirm_prompt, aiohttp_session):
        super().__init__(reference_tentacles_dir, tentacle_path,
                         bot_installation_path, use_confirm_prompt, aiohttp_session)
        self.available_tentacles = []

    async def process(self, name_filter=None) -> int:
        self.available_tentacles = await load_tentacle_with_metadata(self.tentacle_path)
        return await super().process(name_filter)

    def _should_tentacle_be_processed(self, tentacle, name_filter):
        name = tentacle.name
        if name_filter is None or name in name_filter:
            installed_tentacle = Tentacle.find(self.available_tentacles, name)
            if installed_tentacle is not None:
                installed_version = installed_tentacle.version
                return LooseVersion(installed_version) < LooseVersion(tentacle.version)
        return False



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import aiohttp
import pytest
from os import path, walk

from octobot_tentacles_manager.constants import TENTACLE_TYPES, TENTACLES_ARCHIVE_ROOT, DEFAULT_TENTACLES_URL
from octobot_tentacles_manager.util.tentacle_fetching import fetch_and_extract_tentacles, cleanup_temp_dirs

# All test coroutines will be treated as marked.

pytestmark = pytest.mark.asyncio

temp_dir = "tests_temp"


async def test_fetch_and_extract_tentacles_using_download():
    _cleanup()
    # ensure cleanup is working
    assert not path.isdir(temp_dir)
    session = aiohttp.ClientSession()
    await fetch_and_extract_tentacles(temp_dir, DEFAULT_TENTACLES_URL, session)
    _test_temp_tentacles()
    _cleanup()


async def test_fetch_and_extract_tentacles_using_download_without_session():
    _cleanup()
    with pytest.raises(RuntimeError):
        await fetch_and_extract_tentacles(temp_dir, DEFAULT_TENTACLES_URL, None)
    assert not path.isdir(temp_dir)


async def test_fetch_and_extract_tentacles_using_local_file():
    _cleanup()
    await fetch_and_extract_tentacles(temp_dir, path.join("tests", "static", "tentacles.zip"), None)
    _test_temp_tentacles()
    _cleanup()


def _test_temp_tentacles():
    assert all(path.isdir(_tentacle_path(tentacle_type)) for tentacle_type in TENTACLE_TYPES)
    # assert sub directories also got extracted
    total_files_count = sum(1 for _ in walk(temp_dir))
    assert total_files_count > len(TENTACLE_TYPES)


def _tentacle_path(tentacle_type):
    return path.join(temp_dir, TENTACLES_ARCHIVE_ROOT, tentacle_type)


def _cleanup():
    cleanup_temp_dirs(temp_dir)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from asyncio import gather

from octobot_tentacles_manager.managers.tentacle_manager import TentacleManager
from octobot_tentacles_manager.managers.tentacles_init_files_manager import update_tentacle_type_init_file
from octobot_tentacles_manager.workers.tentacles_worker import TentaclesWorker
from octobot_tentacles_manager.models.tentacle import Tentacle
from octobot_tentacles_manager.util.tentacle_explorer import load_tentacle_with_metadata


class InstallWorker(TentaclesWorker):

    async def process(self, name_filter=None) -> int:
        await self.tentacles_setup_manager.create_missing_tentacles_arch()
        self.reset_worker()
        self.progress = 1
        all_tentacles = await load_tentacle_with_metadata(self.reference_tentacles_root)
        self.register_error_on_missing_tentacles(all_tentacles, name_filter)
        to_install_tentacles = [tentacle
                                for tentacle in all_tentacles
                                if self._should_tentacle_be_processed(tentacle, name_filter)]
        self.total_steps = len(to_install_tentacles)
        self.register_to_process_tentacles_modules(to_install_tentacles)
        await gather(*[self._install_tentacle(tentacle) for tentacle in to_install_tentacles])
        await self.tentacles_setup_manager.refresh_user_tentacles_setup_config_file()
        self.tentacles_setup_manager.cleanup_temp_dirs()
        self.log_summary()
        return len(self.errors)

    def _should_tentacle_be_processed(self, tentacle, name_filter):
        return name_filter is None or tentacle.name in name_filter

    async def _install_tentacle(self, tentacle):
        try:
            if tentacle.name not in self.processed_tentacles_modules:
                self.processed_tentacles_modules.append(tentacle.name)
                await self.handle_requirements(tentacle, self._try_install_from_requirements)
                tentacle_manager = TentacleManager(tentacle, self.bot_installation_path)
                await tentacle_manager.install_tentacle(self.tentacle_path)
                update_tentacle_type_init_file(tentacle, tentacle_manager.target_tentacle_path)
                self.logger.info(f"[{self.progress}/{self.total_steps}] installed {tentacle}")
        except Exception as e:
            message = f"Error when installing {tentacle.name}: {e}"
            self.errors.append(message)
            self.logger.exception(e, True, message)
        finally:
            self.progress += 1

    async def _try_install_from_requirements(self, tentacle, missing_requirements):
        for requirement, version in missing_requirements.items():
            if TentacleManager.is_requirement_satisfied(requirement, version, tentacle,
                                                        self.fetched_for_requirements_tentacles_versions):
                to_install_tentacle = Tentacle.find(self.fetched_for_requirements_tentacles, requirement)
                if to_install_tentacle is not None:
                    await self._install_tentacle(to_install_tentacle)
                else:
                    raise RuntimeError(f"Can't find {requirement} tentacle required for {tentacle.name}")



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from os.path import exists, join

from octobot_tentacles_manager.api.installer import repair_installation
from octobot_tentacles_manager.configuration.tentacle_configuration import get_config, \
    factory_reset_config, get_config_schema_path, update_config
from octobot_tentacles_manager.configuration.tentacles_setup_configuration import TentaclesSetupConfiguration
from octobot_tentacles_manager.constants import USER_TENTACLE_CONFIG_FILE_PATH, DEFAULT_BOT_PATH, TENTACLES_PATH


async def ensure_setup_configuration(tentacle_path=TENTACLES_PATH, bot_path=DEFAULT_BOT_PATH) -> None:
    if not exists(join(bot_path, USER_TENTACLE_CONFIG_FILE_PATH)):
        await repair_installation(tentacle_path, bot_path, verbose=False)


async def get_tentacles_setup_config(config_path=USER_TENTACLE_CONFIG_FILE_PATH) -> TentaclesSetupConfiguration:
    setup_config = TentaclesSetupConfiguration(config_path=config_path)
    await setup_config.read_config()
    return setup_config


def is_tentacle_activated_in_tentacles_setup_config(tentacles_setup_config, klass_name, default_value=False,
                                                    raise_errors=False) -> bool:
    try:
        return tentacles_setup_config.tentacles_activation[klass_name]
    except KeyError as e:
        if raise_errors:
            raise e
        return default_value


def get_tentacles_activation(tentacles_setup_config) -> dict:
    return tentacles_setup_config.tentacles_activation


def update_activation_configuration(tentacles_setup_config, new_config, deactivate_others) -> bool:
    return tentacles_setup_config.update_activation_configuration(new_config, deactivate_others)


def get_user_tentacles_packages(tentacles_setup_config) -> dict:
    return tentacles_setup_config.user_tentacles


async def save_tentacles_setup_configuration(tentacles_setup_config) -> None:
    await tentacles_setup_config.save_config()


def get_activated_tentacles(tentacles_setup_config) -> list:
    return [tentacle_class
            for tentacle_class, activated in tentacles_setup_config.tentacles_activation.items()
            if activated]


def update_tentacle_config(tentacle_class: object, config_data: dict) -> None:
    update_config(tentacle_class, config_data)


def get_tentacle_config(klass) -> dict:
    return get_config(klass)


def factory_tentacle_reset_config(klass) -> None:
    return factory_reset_config(klass)


def get_tentacle_config_schema_path(klass) -> str:
    return get_config_schema_path(klass)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from asyncio import gather
from os import scandir, DirEntry
from os.path import isdir, join, sep

from octobot_tentacles_manager.constants import PYTHON_EXT, PYTHON_INIT_FILE, TENTACLE_MAX_SUB_FOLDERS_LEVEL, \
    FOLDERS_BLACK_LIST
from octobot_tentacles_manager.models.tentacle_factory import TentacleFactory
from octobot_tentacles_manager.models.tentacle_type import TentacleType


async def load_tentacle_with_metadata(tentacle_path: str):
    loaded_tentacles = _parse_all_tentacles(tentacle_path)
    await _load_all_metadata(loaded_tentacles)
    return loaded_tentacles


async def _load_all_metadata(tentacles):
    await gather(*[tentacle.initialize() for tentacle in tentacles])


def _parse_all_tentacles(root: str):
    factory = TentacleFactory(root)
    return [factory.create_tentacle_from_type(tentacle_entry.name, tentacle_type)
            for tentacle_type in _get_tentacle_types(root)
            for tentacle_entry in scandir(join(root, tentacle_type.to_path()))
            if not (tentacle_entry.name == PYTHON_INIT_FILE or tentacle_entry.name in FOLDERS_BLACK_LIST)]


def _get_tentacle_types(ref_tentacles_root):
    tentacle_types = []
    if isdir(ref_tentacles_root):
        _find_tentacles_type_in_directories(ref_tentacles_root, tentacle_types)
    return tentacle_types


def _find_tentacles_type_in_directories(ref_tentacles_root: DirEntry, tentacle_types: list, current_level=1):
    if current_level <= TENTACLE_MAX_SUB_FOLDERS_LEVEL:
        for tentacle_type_entry in scandir(ref_tentacles_root):
            if tentacle_type_entry.is_dir():
                _explore_tentacle_dir(tentacle_type_entry, tentacle_types, current_level)


def _explore_tentacle_dir(tentacle_type_entry: DirEntry, tentacle_types: list, nesting_level: int):
    # full_tentacle_type_path is the path to the current directory starting from the most global Tentacle type without
    # anything before it:
    # ex: with a at tentacle_dir="downloaded_tentacles/xyz/Backtesting/importers" then
    # full_tentacle_type_path will be "Backtesting/importers"
    full_tentacle_type_path = tentacle_type_entry.path.split(sep)[-nesting_level:]
    if not _add_tentacle_type_if_is_valid(tentacle_type_entry, join(*full_tentacle_type_path), tentacle_types):
        # no tentacle in this folder, look into nested folders
        _find_tentacles_type_in_directories(tentacle_type_entry, tentacle_types, nesting_level + 1)


def _add_tentacle_type_if_is_valid(tentacle_type_entry: DirEntry, full_tentacle_type_path: str, tentacle_types: list):
    if _has_tentacle_in_direct_sub_directories(tentacle_type_entry):
        tentacle_types.append(TentacleType(full_tentacle_type_path))
        return True
    return False


def _has_tentacle_in_direct_sub_directories(directory_entry: DirEntry):
    return any((file_entry.name.endswith(PYTHON_EXT) and not file_entry.name == PYTHON_INIT_FILE)
               for sub_directory_entry in scandir(directory_entry)
               if sub_directory_entry.is_dir()
               for file_entry in scandir(sub_directory_entry)
               )



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from os import scandir, remove, mkdir
from os.path import exists, join
from shutil import rmtree, copytree, make_archive, copy

from octobot_commons.logging.logging_util import get_logger
from octobot_tentacles_manager.constants import TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER, TENTACLES_ARCHIVE_ROOT, \
    TENTACLES_PACKAGE_FORMAT, PYTHON_GENERATED_ELEMENTS, PYTHON_GENERATED_ELEMENTS_EXTENSION
from octobot_tentacles_manager.managers.tentacles_setup_manager import TentaclesSetupManager
from octobot_tentacles_manager.util.tentacle_explorer import load_tentacle_with_metadata


async def create_tentacles_package_from_local_tentacles(package_name, tentacles_folder, in_zip, with_dev_mode) -> int:
    logger = get_logger("tentacles_package_creator")
    try:
        # create working folder
        working_folder = join(TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER, TENTACLES_ARCHIVE_ROOT) if in_zip else package_name
        if in_zip:
            _create_zip_working_tentacles_folder(tentacles_folder, working_folder)
        else:
            _create_folder_working_tentacles_folder(tentacles_folder, working_folder)

        # remove dev-mode tentacles if necessary
        if not with_dev_mode:
            await _cleanup_in_dev_tentacles(working_folder)

        # cleanup temp working folder
        tentacles_setup_manager = TentaclesSetupManager(working_folder)
        await tentacles_setup_manager.remove_tentacle_arch_init_files()
        _remove_python_generated_files(working_folder)

        if in_zip:
            _zip_tentacles_package(package_name, TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER)
            logger.info(f"Zipped tentacles package available at: {package_name}")
        else:
            logger.info(f"Cleaned tentacles folder available at: {package_name}")

        return 0
    except Exception as e:
        logger.exception(e, True, f"Error while creating tentacles archive: {e}")
        return 1


def _zip_tentacles_package(package_name, working_folder):
    # remove .zip extension if necessary
    file_name = package_name.split(f".{TENTACLES_PACKAGE_FORMAT}")[0]
    make_archive(file_name, TENTACLES_PACKAGE_FORMAT, working_folder)
    # remove working folder
    rmtree(TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER)


async def _cleanup_in_dev_tentacles(tentacles_folder):
    for tentacle in await load_tentacle_with_metadata(tentacles_folder):
        if tentacle.in_dev_mode:
            rmtree(join(tentacle.tentacle_path, tentacle.name))


def _create_zip_working_tentacles_folder(source_tentacles_folder, working_folder):
    if exists(TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER):
        rmtree(TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER)
    mkdir(TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER)
    copytree(source_tentacles_folder, working_folder)


def _create_folder_working_tentacles_folder(source_tentacles_folder, working_folder):
    if not exists(working_folder):
        copytree(source_tentacles_folder, working_folder)
    else:
        _merge_folders(source_tentacles_folder, working_folder)


def _merge_folders(to_merge_folder, dest_folder):
    dest_folder_elements = {
        element.name: element for element in scandir(dest_folder)
    }
    for element in scandir(to_merge_folder):
        dest = join(dest_folder, element.name)
        if element.is_file():
            copy(element.path, dest)
        else:
            if element.name not in dest_folder_elements:
                copytree(element.path, dest)
            else:
                _merge_folders(element.path, dest_folder_elements[element.name].path)


def _remove_python_generated_files(directory):
    for element in scandir(directory):
        element_ext = element.name.split(".")[-1]
        if element.name in PYTHON_GENERATED_ELEMENTS or \
                (element_ext in PYTHON_GENERATED_ELEMENTS_EXTENSION and element.is_file()):
            if element.is_dir():
                rmtree(element)
            elif element.is_file():
                remove(element)
        elif element.is_dir():
            _remove_python_generated_files(element)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import aiofiles
from os import path, makedirs


async def find_or_create(path_to_create, is_directory=True, file_content=""):
    if not path.exists(path_to_create):
        if is_directory:
            if not path.isdir(path_to_create):
                makedirs(path_to_create)
        else:
            if not path.isfile(path_to_create):
                # should be used for python init.py files only
                async with aiofiles.open(path_to_create, "w+") as file:
                    await file.write(file_content)
        return True
    return False



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_tentacles_manager.constants import TENTACLES_PATH, DEFAULT_BOT_PATH
from octobot_tentacles_manager.api.util.tentacles_management import manage_tentacles
from octobot_tentacles_manager.workers.uninstall_worker import UninstallWorker


USER_HELP = """Uninstall the given tentacle modules. 
    Does not delete the associated tentacle configuration."""


async def uninstall_all_tentacles(tentacle_path=TENTACLES_PATH,
                                  bot_path=DEFAULT_BOT_PATH, use_confirm_prompt=False) -> int:
    return await _uninstall_tentacles(None, tentacle_path, bot_path, use_confirm_prompt)


async def uninstall_tentacles(tentacle_names,
                              tentacle_path=TENTACLES_PATH,
                              bot_path=DEFAULT_BOT_PATH,
                              use_confirm_prompt=False) -> int:
    return await _uninstall_tentacles(tentacle_names, tentacle_path, bot_path, use_confirm_prompt)


async def _uninstall_tentacles(tentacle_names,
                               tentacle_path=TENTACLES_PATH,
                               bot_path=DEFAULT_BOT_PATH,
                               use_confirm_prompt=False) -> int:
    uninstall_worker = UninstallWorker(None, tentacle_path, bot_path, use_confirm_prompt, None)
    return await manage_tentacles(uninstall_worker, tentacle_names)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from os import scandir
from os.path import join, exists, isdir
from shutil import copyfile, rmtree, copytree

from octobot_commons.logging.logging_util import get_logger
from octobot_tentacles_manager.constants import USER_TENTACLE_SPECIFIC_CONFIG_PATH, CONFIG_SCHEMA_EXT, \
    TENTACLE_CONFIG, CONFIG_EXT, DEFAULT_BOT_PATH
from octobot_tentacles_manager.managers.tentacles_init_files_manager import create_tentacle_init_file_if_necessary
from octobot_tentacles_manager.util.file_util import find_or_create


class TentacleManager:

    def __init__(self, tentacle, bot_installation_path=DEFAULT_BOT_PATH):
        self.tentacle = tentacle
        self.bot_installation_path = bot_installation_path
        self.target_tentacle_path = None

    async def install_tentacle(self, tentacle_path):
        self.target_tentacle_path = join(tentacle_path, self.tentacle.tentacle_type.to_path())
        tentacle_module_path = join(self.target_tentacle_path, self.tentacle.name)
        await self._update_tentacle_folder(tentacle_module_path)
        await create_tentacle_init_file_if_necessary(tentacle_module_path, self.tentacle)
        self.import_tentacle_config_if_any(tentacle_module_path)

    async def uninstall_tentacle(self):
        rmtree(join(self.bot_installation_path, self.tentacle.tentacle_path, self.tentacle.name))

    @staticmethod
    def find_tentacles_missing_requirements(tentacle, version_by_modules):
        # check if requirement is in tentacles to be installed in this call
        return {
            requirement: version
            for requirement, version in tentacle.extract_tentacle_requirements()
            if not TentacleManager.is_requirement_satisfied(requirement, version, tentacle, version_by_modules)
        }

    @staticmethod
    def is_requirement_satisfied(requirement, version, tentacle, version_by_modules):
        satisfied = False
        if requirement in version_by_modules:
            available = version_by_modules[requirement]
            if version is None:
                satisfied = True
            elif version != available:
                get_logger(TentacleManager.__name__).\
                    error(f"Incompatible tentacle version requirement for "
                          f"{tentacle.name}: requires {version}, installed: "
                          f"{available}. This tentacle might not work as expected")
                satisfied = True
        return satisfied

    async def _update_tentacle_folder(self, target_tentacle_path):
        reference_tentacle_path = join(self.tentacle.tentacle_path, self.tentacle.name)
        await find_or_create(target_tentacle_path)
        for tentacle_file_entry in scandir(reference_tentacle_path):
            target_file_or_dir = join(target_tentacle_path, tentacle_file_entry.name)
            if tentacle_file_entry.is_file():
                copyfile(tentacle_file_entry, target_file_or_dir)
            else:
                if exists(target_file_or_dir):
                    rmtree(target_file_or_dir)
                copytree(tentacle_file_entry, target_file_or_dir)

    def import_tentacle_config_if_any(self, tentacle_module_path, replace=False):
        target_tentacle_config_path = join(tentacle_module_path, TENTACLE_CONFIG)
        if isdir(target_tentacle_config_path):
            for config_file_entry in scandir(target_tentacle_config_path):
                if config_file_entry.name.endswith(CONFIG_EXT) and not config_file_entry.name.endswith(CONFIG_SCHEMA_EXT):
                    target_user_path = \
                        join(self.bot_installation_path, USER_TENTACLE_SPECIFIC_CONFIG_PATH, config_file_entry.name)
                    if replace or not exists(target_user_path):
                        copyfile(join(target_tentacle_config_path, config_file_entry.name), target_user_path)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_tentacles_manager.api.util.tentacles_management import manage_tentacles
from octobot_tentacles_manager.constants import TENTACLES_INSTALL_TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH
from octobot_tentacles_manager.workers.install_worker import InstallWorker
from octobot_tentacles_manager.workers.repair_worker import RepairWorker
from octobot_tentacles_manager.workers.single_install_worker import SingleInstallWorker

USER_HELP = """Install or re-install the given tentacles modules with their requirements if any.
    Does not edit tentacles configuration files and creates it if missing."""


async def install_all_tentacles(tentacles_path_or_url, tentacle_path=TENTACLES_PATH,
                                bot_path=DEFAULT_BOT_PATH, aiohttp_session=None) -> int:
    return await _install_tentacles(None, tentacles_path_or_url,
                                    tentacle_path,
                                    bot_path,
                                    aiohttp_session=aiohttp_session)


async def install_tentacles(tentacle_names, tentacles_path_or_url,
                            bot_path=DEFAULT_BOT_PATH, tentacle_path=TENTACLES_PATH, aiohttp_session=None) -> int:
    return await _install_tentacles(tentacle_names, tentacles_path_or_url,
                                    tentacle_path, bot_path, aiohttp_session=aiohttp_session)


async def install_single_tentacle(single_tentacle_path, single_tentacle_type, tentacle_path=TENTACLES_PATH,
                                  bot_path=DEFAULT_BOT_PATH, aiohttp_session=None) -> int:
    single_install_worker = SingleInstallWorker(TENTACLES_INSTALL_TEMP_DIR, tentacle_path,
                                                bot_path, False, aiohttp_session)
    single_install_worker.single_tentacle_path = single_tentacle_path
    single_install_worker.single_tentacle_type = single_tentacle_type
    return await manage_tentacles(single_install_worker, None, None, aiohttp_session)


async def repair_installation(tentacle_path=TENTACLES_PATH, bot_path=DEFAULT_BOT_PATH, verbose=True) -> int:
    repair_worker = RepairWorker(None, tentacle_path, bot_path, False, None)
    repair_worker.verbose = verbose
    return await manage_tentacles(repair_worker, None, None, None)


async def _install_tentacles(tentacle_names, tentacles_path_or_url, tentacle_path=TENTACLES_PATH,
                             bot_path=DEFAULT_BOT_PATH, use_confirm_prompt=False, aiohttp_session=None) -> int:
    install_worker = InstallWorker(TENTACLES_INSTALL_TEMP_DIR, tentacle_path,
                                   bot_path, use_confirm_prompt, aiohttp_session)
    install_worker.tentacles_path_or_url = tentacles_path_or_url
    return await manage_tentacles(install_worker, tentacle_names, tentacles_path_or_url, aiohttp_session)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_tentacles_manager.models.tentacle import Tentacle
from octobot_tentacles_manager.models.tentacle_type import TentacleType


class TentacleFactory:
    def __init__(self, tentacle_root_path):
        self.tentacle_root_path = tentacle_root_path

    def create_tentacle_from_type(self, name, tentacle_type):
        return Tentacle(self.tentacle_root_path, name, tentacle_type)

    async def create_and_load_tentacle_from_module(self, tentacle_module):
        tentacle_type = TentacleType.from_import_path(self.tentacle_root_path, tentacle_module.__name__)
        tentacle = Tentacle(self.tentacle_root_path, tentacle_type.module_name, tentacle_type)
        await tentacle.initialize()
        return tentacle
